# ruff: noqa: ARG002
from typing import Optional

from xgboost.callback import TrainingCallback

from dvclive import Live


class DVCLiveCallback(TrainingCallback):
    def __init__(
        self, metric_data, model_file=None, live: Optional[Live] = None, **kwargs
    ):
        super().__init__()
        self._metric_data = metric_data
        self.model_file = model_file
        self.live = live if live is not None else Live(**kwargs)

    def after_iteration(self, model, epoch, evals_log):
        for key, values in evals_log[self._metric_data].items():
            if values:
                latest_metric = values[-1]
            self.live.log_metric(key, latest_metric)
        if self.model_file:
            model.save_model(self.model_file)
        self.live.next_step()

    def after_training(self, model):
        self.live.end()
        return model



# ruff: noqa: ARG002
import os
from typing import Dict, Optional

from tensorflow.keras.callbacks import Callback
from tensorflow.keras.models import load_model

from dvclive import Live
from dvclive.utils import standardize_metric_name


class DVCLiveCallback(Callback):
    def __init__(
        self,
        model_file=None,
        save_weights_only: bool = False,
        live: Optional[Live] = None,
        **kwargs,
    ):
        super().__init__()
        self.model_file = model_file
        self.save_weights_only = save_weights_only
        self.live = live if live is not None else Live(**kwargs)

    def on_train_begin(self, logs=None):
        if (
            self.live._resume  # noqa: SLF001
            and self.model_file is not None
            and os.path.exists(self.model_file)
        ):
            if self.save_weights_only:
                self.model.load_weights(self.model_file)
            else:
                self.model = load_model(self.model_file)

    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):
        logs = logs or {}
        for metric, value in logs.items():
            self.live.log_metric(standardize_metric_name(metric, __name__), value)
        if self.model_file:
            if self.save_weights_only:
                self.model.save_weights(self.model_file)
            else:
                self.model.save(self.model_file)
            self.live.log_artifact(self.model_file)
        self.live.next_step()

    def on_train_end(self, logs: Optional[Dict] = None):
        self.live.end()



# ruff: noqa: SLF001
import copy
import os
from pathlib import Path

from dvclive import env
from dvclive.plots import Image, Metric
from dvclive.serialize import dump_yaml

_CHECKPOINT_SLEEP = 0.1


def _dvc_dir(dirname):
    return os.path.join(dirname, ".dvc")


def _dvc_exps_run_dir(dirname: str) -> str:
    return os.path.join(dirname, ".dvc", "tmp", "exps", "run")


def _dvclive_only_signal_file(root_dir: str) -> str:
    dvc_exps_run_dir = _dvc_exps_run_dir(root_dir)
    return os.path.join(dvc_exps_run_dir, "DVCLIVE_ONLY")


def _find_dvc_root(root=None):
    if not root:
        root = os.getcwd()

    root = os.path.realpath(root)

    if not os.path.isdir(root):
        raise NotADirectoryError(f"'{root}'")

    while True:
        if os.path.exists(_dvc_dir(root)):
            return root
        if os.path.ismount(root):
            break
        root = os.path.dirname(root)

    return None


def _write_file(file: str, contents=""):
    import builtins

    with builtins.open(file, "w", encoding="utf-8") as fobj:
        # NOTE: force flushing/writing empty file to disk, otherwise when
        # run in certain contexts (pytest) file may not actually be written
        fobj.write(str(contents))
        fobj.flush()
        os.fsync(fobj.fileno())


def make_checkpoint():
    from time import sleep

    root_dir = _find_dvc_root()
    if not root_dir:
        return

    signal_file = os.path.join(root_dir, ".dvc", "tmp", env.DVC_CHECKPOINT)

    _write_file(signal_file)

    while os.path.exists(signal_file):
        sleep(_CHECKPOINT_SLEEP)


def get_dvc_repo():
    from dvc.exceptions import NotDvcRepoError
    from dvc.repo import Repo
    from dvc.scm import Git, SCMError
    from scmrepo.exceptions import SCMError as GitSCMError

    try:
        return Repo()
    except (NotDvcRepoError, SCMError):
        try:
            return Repo.init(Git().root_dir)
        except GitSCMError:
            return None


def make_dvcyaml(live):
    dvcyaml = {}
    if live._params:
        dvcyaml["params"] = [os.path.relpath(live.params_file, live.dir)]
    if live._metrics or live.summary:
        dvcyaml["metrics"] = [os.path.relpath(live.metrics_file, live.dir)]
    plots = []
    plots_path = Path(live.plots_dir)
    if live._metrics:
        metrics_path = (plots_path / Metric.subfolder).relative_to(live.dir).as_posix()
        metrics_config = {metrics_path: {"x": "step"}}
        plots.append(metrics_config)
    if live._images:
        images_path = (plots_path / Image.subfolder).relative_to(live.dir)
        plots.append(images_path.as_posix())
    if live._plots:
        for plot in live._plots.values():
            plot_path = plot.output_path.relative_to(live.dir)
            plots.append({plot_path.as_posix(): plot.plot_config})
    if plots:
        dvcyaml["plots"] = plots

    if live._artifacts:
        dvcyaml["artifacts"] = copy.deepcopy(live._artifacts)
        for artifact in dvcyaml["artifacts"].values():
            abs_path = os.path.abspath(artifact["path"])
            abs_dir = os.path.realpath(live.dir)
            relative_path = os.path.relpath(abs_path, abs_dir)
            artifact["path"] = Path(relative_path).as_posix()

    dump_yaml(dvcyaml, live.dvc_file)


def mark_dvclive_only_started():
    """
    Signal DVC VS Code extension that
    an experiment is running in the workspace.
    """
    root_dir = _find_dvc_root()
    if not root_dir:
        return

    exp_run_dir = _dvc_exps_run_dir(root_dir)
    os.makedirs(exp_run_dir, exist_ok=True)

    signal_file = _dvclive_only_signal_file(root_dir)

    _write_file(signal_file, os.getpid())


def mark_dvclive_only_ended():
    root_dir = _find_dvc_root()
    if not root_dir:
        return

    signal_file = _dvclive_only_signal_file(root_dir)

    if not os.path.exists(signal_file):
        return

    os.remove(signal_file)


def get_random_exp_name(scm, baseline_rev):
    from dvc.repo.experiments.utils import (
        get_random_exp_name as dvc_get_random_exp_name,
    )

    return dvc_get_random_exp_name(scm, baseline_rev)



import csv
import json
import os
import re
import shutil
import webbrowser
from pathlib import Path
from platform import uname
from typing import Union

StrPath = Union[str, Path]


def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
        return None

    wrapper.has_run = False
    return wrapper


@run_once
def open_file_in_browser(file) -> bool:
    path = Path(file)
    url = str(path) if "Microsoft" in uname().release else path.resolve().as_uri()

    return webbrowser.open(url)


def env2bool(var, undefined=False):
    """
    undefined: return value if env var is unset
    """
    var = os.getenv(var, None)
    if var is None:
        return undefined
    return bool(re.search("1|y|yes|true", var, flags=re.I))


def standardize_metric_name(metric_name: str, framework: str) -> str:
    """Map framework-specific format to DVCLive standard.

    Use `{split}/` as prefix in order to separate by subfolders.
    Use `{train|eval}` as split name.
    """
    if framework == "dvclive.fastai":
        metric_name = metric_name.replace("train_", "train/")
        metric_name = metric_name.replace("valid_", "eval/")

    elif framework == "dvclive.huggingface":
        for split in {"train", "eval"}:
            metric_name = metric_name.replace(f"{split}_", f"{split}/")

    elif framework == "dvclive.keras":
        if "val_" in metric_name:
            metric_name = metric_name.replace("val_", "eval/")
        else:
            metric_name = f"train/{metric_name}"

    elif framework == "dvclive.lightning":
        parts = metric_name.split("_")
        split, freq, rest = None, None, None
        if parts[0] in ["train", "val", "test"]:
            split = parts.pop(0)
            # Only set freq if split was also found.
            # Otherwise we end up conflicting with out internal `step` property.
            if parts[-1] in ["step", "epoch"]:
                freq = parts.pop()
        rest = "_".join(parts)
        parts = [part for part in (split, freq, rest) if part]
        metric_name = "/".join(parts)

    return metric_name


def parse_tsv(path):
    with open(path, encoding="utf-8", newline="") as fd:
        reader = csv.DictReader(fd, delimiter="\t")
        return list(reader)


def parse_json(path):
    with open(path, encoding="utf-8") as fd:
        return json.load(fd)


def parse_metrics(live):
    from .plots import Metric

    metrics_path = Path(live.plots_dir) / Metric.subfolder
    history = {}
    for suffix in Metric.suffixes:
        for scalar_file in metrics_path.rglob(f"*{suffix}"):
            history[str(scalar_file)] = parse_tsv(scalar_file)
    latest = parse_json(live.metrics_file)
    return history, latest


def matplotlib_installed() -> bool:
    try:
        import matplotlib  # noqa: F401
    except ImportError:
        return False
    return True


def inside_colab() -> bool:
    try:
        from google import colab  # noqa: F401
    except ImportError:
        return False
    return True


def inside_notebook() -> bool:
    if inside_colab():
        return True

    try:
        shell = get_ipython().__class__.__name__  # type: ignore[name-defined]
    except NameError:
        return False

    if shell == "ZMQInteractiveShell":
        import IPython

        return IPython.__version__ >= "6.0.0"
    return False


def clean_and_copy_into(src: StrPath, dst: StrPath) -> str:
    Path(dst).mkdir(exist_ok=True)

    basename = os.path.basename(os.path.normpath(src))
    dst_path = Path(os.path.join(dst, basename))

    if dst_path.is_file() or dst_path.is_symlink():
        dst_path.unlink()
    elif dst_path.is_dir():
        shutil.rmtree(dst_path)

    if os.path.isdir(src):
        shutil.copytree(src, dst_path)
    else:
        shutil.copy2(src, dst_path)

    return str(dst_path)



import pytest
from IPython import display
from PIL import Image

from dvclive import Live
from dvclive.env import DVCLIVE_OPEN
from dvclive.error import InvalidReportModeError
from dvclive.plots import CustomPlot
from dvclive.plots import Image as LiveImage
from dvclive.plots import Metric
from dvclive.plots.sklearn import SKLearnPlot
from dvclive.report import (
    get_custom_plot_renderers,
    get_image_renderers,
    get_metrics_renderers,
    get_params_renderers,
    get_scalar_renderers,
    get_sklearn_plot_renderers,
)


@pytest.mark.parametrize("mode", ["html", "md", "notebook"])
def test_get_image_renderers(tmp_dir, mode):
    with Live() as live:
        img = Image.new("RGB", (10, 10), (255, 0, 0))
        live.log_image("image.png", img)

    image_renderers = get_image_renderers(
        tmp_dir / live.plots_dir / LiveImage.subfolder
    )
    assert len(image_renderers) == 1
    img = image_renderers[0].datapoints[0]
    assert img["src"].startswith("data:image;base64,")
    assert img["rev"] == "image.png"


def test_get_renderers(tmp_dir, mocker):
    live = Live()

    live.log_param("string", "goo")
    live.log_param("number", 2)

    for i in range(2):
        live.log_metric("foo/bar", i)
        live.next_step()

    scalar_renderers = get_scalar_renderers(tmp_dir / live.plots_dir / Metric.subfolder)
    assert len(scalar_renderers) == 1
    assert scalar_renderers[0].datapoints == [
        {
            "bar": "0",
            "rev": "workspace",
            "step": "0",
        },
        {
            "bar": "1",
            "rev": "workspace",
            "step": "1",
        },
    ]
    assert scalar_renderers[0].properties["y"] == "bar"
    assert scalar_renderers[0].properties["title"] == "foo/bar"
    assert scalar_renderers[0].name == "static/foo/bar"

    metrics_renderer = get_metrics_renderers(live.metrics_file)[0]
    assert metrics_renderer.datapoints == [{"step": 1, "foo": {"bar": 1}}]

    params_renderer = get_params_renderers(live.params_file)[0]
    assert params_renderer.datapoints == [{"string": "goo", "number": 2}]


def test_report_init(monkeypatch, mocker):
    monkeypatch.setenv("CI", "false")
    live = Live()
    assert live._report_mode == "html"

    monkeypatch.setenv("CI", "true")
    live = Live()
    assert live._report_mode == "md"

    mocker.patch("dvclive.live.matplotlib_installed", return_value=False)
    live = Live()
    assert live._report_mode == "html"

    for report in {None, "html", "md"}:
        live = Live(report=report)
        assert live._report_mode == report

    with pytest.raises(InvalidReportModeError, match="Got foo instead."):
        Live(report="foo")


@pytest.mark.parametrize("mode", ["html", "md"])
def test_make_report(tmp_dir, mode):
    last_report = ""
    live = Live(report=mode)
    for i in range(3):
        live.log_metric("foobar", i)
        live.log_metric("foo/bar", i)
        live.make_report()
        live.next_step()
        assert (tmp_dir / live.report_file).exists()
        current_report = (tmp_dir / live.report_file).read_text()
        assert last_report != current_report
        last_report = current_report


@pytest.mark.vscode()
def test_make_report_open(tmp_dir, mocker, monkeypatch):
    mocked_open = mocker.patch("webbrowser.open")
    live = Live()
    live.log_sklearn_plot("confusion_matrix", [0, 0, 1, 1], [1, 0, 0, 1])
    live.make_report()
    live.make_report()

    assert not mocked_open.called

    live = Live(report=None)
    live.log_metric("foo", 1)
    live.next_step()

    assert not mocked_open.called

    monkeypatch.setenv(DVCLIVE_OPEN, True)

    live = Live()
    live.log_sklearn_plot("confusion_matrix", [0, 0, 1, 1], [1, 0, 0, 1])
    live.make_report()

    mocked_open.assert_called_once()


def test_get_plot_renderers_sklearn(tmp_dir):
    live = Live()

    for _ in range(2):
        live.log_sklearn_plot("confusion_matrix", [0, 0, 1, 1], [1, 0, 0, 1])
        live.log_sklearn_plot(
            "confusion_matrix", [0, 0, 1, 1], [1, 0, 0, 1], name="train/cm"
        )
        live.log_sklearn_plot("roc", [0, 0, 1, 1], [1, 0.1, 0, 1], name="roc_curve")
        live.log_sklearn_plot(
            "roc", [0, 0, 1, 1], [1, 0.1, 0, 1], name="other_roc.json"
        )
        live.next_step()

    plot_renderers = get_sklearn_plot_renderers(
        tmp_dir / live.plots_dir / SKLearnPlot.subfolder, live
    )
    assert len(plot_renderers) == 4
    plot_renderers_dict = {
        plot_renderer.name: plot_renderer for plot_renderer in plot_renderers
    }
    for name in ("roc_curve", "other_roc"):
        plot_renderer = plot_renderers_dict[name]
        assert plot_renderer.datapoints == [
            {"fpr": 0.0, "rev": "workspace", "threshold": 2.0, "tpr": 0.0},
            {"fpr": 0.5, "rev": "workspace", "threshold": 1.0, "tpr": 0.5},
            {"fpr": 1.0, "rev": "workspace", "threshold": 0.1, "tpr": 0.5},
            {"fpr": 1.0, "rev": "workspace", "threshold": 0.0, "tpr": 1.0},
        ]
        assert plot_renderer.properties == live._plots[name].plot_config

    for name in ("confusion_matrix", "train/cm"):
        plot_renderer = plot_renderers_dict[name]
        assert plot_renderer.datapoints == [
            {"actual": "0", "rev": "workspace", "predicted": "1"},
            {"actual": "0", "rev": "workspace", "predicted": "0"},
            {"actual": "1", "rev": "workspace", "predicted": "0"},
            {"actual": "1", "rev": "workspace", "predicted": "1"},
        ]
        assert plot_renderer.properties == live._plots[name].plot_config


def test_get_plot_renderers_custom(tmp_dir):
    live = Live()

    datapoints = [{"x": 1, "y": 2}, {"x": 3, "y": 4}]
    for _ in range(2):
        live.log_plot("foo_default", datapoints, x="x", y="y")
        live.log_plot(
            "foo_scatter",
            datapoints,
            x="x",
            y="y",
            template="scatter",
        )
        live.next_step()
    plot_renderers = get_custom_plot_renderers(
        tmp_dir / live.plots_dir / CustomPlot.subfolder, live
    )

    assert len(plot_renderers) == 2
    plot_renderers_dict = {
        plot_renderer.name: plot_renderer for plot_renderer in plot_renderers
    }
    for name in ("foo_default", "foo_scatter"):
        plot_renderer = plot_renderers_dict[name]
        assert plot_renderer.datapoints == [
            {"rev": "workspace", "x": 1, "y": 2},
            {"rev": "workspace", "x": 3, "y": 4},
        ]
        assert plot_renderer.properties == live._plots[name].plot_config


def test_report_auto_doesnt_set_notebook(tmp_dir, mocker):
    mocker.patch("dvclive.live.inside_notebook", return_value=True)
    live = Live()
    assert live._report_mode != "notebook"


def test_report_notebook_fallsback_to_html(tmp_dir, mocker):
    mocker.patch("dvclive.live.inside_notebook", return_value=False)
    spy = mocker.spy(display, "display")
    live = Live(report="notebook")
    assert live._report_mode == "html"
    assert not spy.called


def test_report_notebook(tmp_dir, mocker):
    mocker.patch("dvclive.live.inside_notebook", return_value=True)
    mocked_display = mocker.MagicMock()
    mocker.patch("IPython.display.display", return_value=mocked_display)
    live = Live(report="notebook")
    assert live._report_mode == "notebook"
    live.make_report()
    assert mocked_display.update.called



# ruff: noqa: ARG002
import sys

import pytest
from dvc_studio_client.env import DVC_STUDIO_TOKEN, DVC_STUDIO_URL, STUDIO_REPO_URL


@pytest.fixture()
def tmp_dir(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    return tmp_path


@pytest.fixture()
def mocked_dvc_repo(tmp_dir, mocker):
    _dvc_repo = mocker.MagicMock()
    _dvc_repo.index.stages = []
    _dvc_repo.scm.get_rev.return_value = "f" * 40
    _dvc_repo.scm.get_ref.return_value = None
    _dvc_repo.scm.no_commits = False
    _dvc_repo.experiments.save.return_value = "e" * 40
    _dvc_repo.root_dir = tmp_dir
    _dvc_repo.config = {}
    mocker.patch("dvclive.live.get_dvc_repo", return_value=_dvc_repo)
    return _dvc_repo


@pytest.fixture()
def dvc_repo(tmp_dir):
    from dvc.repo import Repo
    from scmrepo.git import Git

    Git.init(tmp_dir)
    repo = Repo.init(tmp_dir)
    repo.scm.add_commit(".", "init")
    return repo


@pytest.fixture(autouse=True)
def _capture_wrap():
    # https://github.com/pytest-dev/pytest/issues/5502#issuecomment-678368525
    sys.stderr.close = lambda *args: None
    sys.stdout.close = lambda *args: None


@pytest.fixture(autouse=True)
def _mocked_webbrowser_open(mocker):
    mocker.patch("webbrowser.open")


@pytest.fixture(autouse=True)
def _mocked_ci(monkeypatch):
    monkeypatch.setenv("CI", "false")


@pytest.fixture()
def mocked_studio_post(mocker, monkeypatch):
    valid_response = mocker.MagicMock()
    valid_response.status_code = 200
    mocked_post = mocker.patch("requests.post", return_value=valid_response)
    monkeypatch.setenv(DVC_STUDIO_URL, "https://0.0.0.0")
    monkeypatch.setenv(STUDIO_REPO_URL, "STUDIO_REPO_URL")
    monkeypatch.setenv(DVC_STUDIO_TOKEN, "STUDIO_TOKEN")
    return mocked_post, valid_response



# ruff: noqa: SLF001
import base64
import json
from pathlib import Path
from typing import TYPE_CHECKING

from dvc_render.html import render_html
from dvc_render.image import ImageRenderer
from dvc_render.markdown import render_markdown
from dvc_render.table import TableRenderer
from dvc_render.vega import VegaRenderer

from dvclive.error import InvalidReportModeError
from dvclive.plots import SKLEARN_PLOTS, CustomPlot, Image, Metric
from dvclive.plots.sklearn import SKLearnPlot
from dvclive.serialize import load_yaml
from dvclive.utils import parse_tsv

if TYPE_CHECKING:
    from dvclive import Live


BLANK_NOTEBOOK_REPORT = """
<div style="width: 100%;height: 700px;text-align: center">
DVCLive Report
</div>
"""


def get_scalar_renderers(metrics_path):
    renderers = []
    for suffix in Metric.suffixes:
        for file in metrics_path.rglob(f"*{suffix}"):
            data = parse_tsv(file)
            for row in data:
                row["rev"] = "workspace"

            name = file.relative_to(metrics_path.parent).with_suffix("")
            name = name.as_posix()
            title = name.replace(metrics_path.name, "").strip("/")
            name = name.replace(metrics_path.name, "static")

            properties = {"x": "step", "y": file.stem, "title": title}
            renderers.append(VegaRenderer(data, name, **properties))
    return renderers


def get_image_renderers(images_folder):
    renderers = []
    for suffix in Image.suffixes:
        all_images = Path(images_folder).rglob(f"*{suffix}")
        for file in sorted(all_images):
            base64_str = base64.b64encode(file.read_bytes()).decode()
            src = f"data:image;base64,{base64_str}"
            name = str(file.relative_to(images_folder))
            data = [
                {
                    ImageRenderer.SRC_FIELD: src,
                    ImageRenderer.TITLE_FIELD: name,
                }
            ]
            renderers.append(ImageRenderer(data, name))
    return renderers


def get_custom_plot_renderers(plots_folder, live):
    renderers = []
    for suffix in CustomPlot.suffixes:
        for file in Path(plots_folder).rglob(f"*{suffix}"):
            name = file.relative_to(plots_folder).with_suffix("").as_posix()

            logged_plot = live._plots[name]
            properties = logged_plot.plot_config

            data = json.loads(file.read_text())

            for row in data:
                row["rev"] = "workspace"

            renderers.append(VegaRenderer(data, name, **properties))
    return renderers


def get_sklearn_plot_renderers(plots_folder, live):
    renderers = []
    for suffix in SKLearnPlot.suffixes:
        for file in Path(plots_folder).rglob(f"*{suffix}"):
            name = file.relative_to(plots_folder).with_suffix("").as_posix()
            properties = {}

            logged_plot = live._plots[name]
            for default_name, plot_class in SKLEARN_PLOTS.items():
                if isinstance(logged_plot, plot_class):
                    properties = logged_plot.plot_config
                    data_field = default_name
                    break

            data = json.loads(file.read_text())

            if data_field in data:
                data = data[data_field]

            for row in data:
                row["rev"] = "workspace"

            renderers.append(VegaRenderer(data, name, **properties))
    return renderers


def get_metrics_renderers(dvclive_summary):
    metrics_path = Path(dvclive_summary)
    if metrics_path.exists():
        return [
            TableRenderer(
                [json.loads(metrics_path.read_text(encoding="utf-8"))],
                metrics_path.name,
            )
        ]
    return []


def get_params_renderers(dvclive_params):
    params_path = Path(dvclive_params)
    if params_path.exists():
        return [
            TableRenderer(
                [load_yaml(params_path)],
                params_path.name,
            )
        ]
    return []


def make_report(live: "Live"):
    plots_path = Path(live.plots_dir)

    renderers = []
    renderers.extend(get_params_renderers(live.params_file))
    renderers.extend(get_metrics_renderers(live.metrics_file))
    renderers.extend(get_scalar_renderers(plots_path / Metric.subfolder))
    renderers.extend(get_image_renderers(plots_path / Image.subfolder))
    renderers.extend(
        get_sklearn_plot_renderers(plots_path / SKLearnPlot.subfolder, live)
    )
    renderers.extend(get_custom_plot_renderers(plots_path / CustomPlot.subfolder, live))

    if live._report_mode == "html":
        render_html(renderers, live.report_file, refresh_seconds=5)
    elif live._report_mode == "notebook":
        from IPython.display import Markdown

        md = render_markdown(renderers)
        if live._report_notebook is not None:
            new_report = Markdown(md)  # type: ignore [assignment]
            live._report_notebook.update(new_report)
    elif live._report_mode == "md":
        render_markdown(renderers, live.report_file)
    else:
        raise InvalidReportModeError(live._report_mode)



import os

import pytest

from dvclive import Live
from dvclive.plots.metric import Metric
from dvclive.utils import parse_metrics

try:
    from dvclive.keras import DVCLiveCallback
except ImportError:
    pytest.skip("skipping keras tests", allow_module_level=True)


@pytest.fixture()
def xor_model():
    import numpy as np
    from tensorflow.python.keras import Sequential
    from tensorflow.python.keras.layers import Activation, Dense

    def make():
        x = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
        y = np.array([[0], [1], [1], [0]])

        model = Sequential()
        model.add(Dense(8, input_dim=2))
        model.add(Activation("relu"))
        model.add(Dense(1))
        model.add(Activation("sigmoid"))

        model.compile(loss="binary_crossentropy", optimizer="sgd", metrics=["accuracy"])

        return model, x, y

    return make


def test_keras_callback(tmp_dir, xor_model, mocker):
    model, x, y = xor_model()

    callback = DVCLiveCallback()
    live = callback.live
    spy = mocker.spy(live, "end")
    model.fit(
        x,
        y,
        epochs=1,
        batch_size=1,
        validation_split=0.2,
        callbacks=[callback],
    )
    spy.assert_called_once()

    assert os.path.exists("dvclive")
    logs, _ = parse_metrics(callback.live)

    scalars = os.path.join(callback.live.plots_dir, Metric.subfolder)
    assert os.path.join(scalars, "train", "accuracy.tsv") in logs
    assert os.path.join(scalars, "eval", "accuracy.tsv") in logs


def test_keras_callback_pass_logger():
    logger = Live("train_logs")

    assert DVCLiveCallback().live is not logger
    assert DVCLiveCallback(live=logger).live is logger


@pytest.mark.parametrize("save_weights_only", [True, False])
def test_keras_model_file(tmp_dir, xor_model, mocker, save_weights_only):
    model, x, y = xor_model()
    save = mocker.spy(model, "save")
    save_weights = mocker.spy(model, "save_weights")

    live_callback = DVCLiveCallback(
        model_file="model.h5", save_weights_only=save_weights_only
    )
    log_artifact = mocker.patch.object(live_callback.live, "log_artifact")
    model.fit(
        x,
        y,
        epochs=1,
        batch_size=1,
        callbacks=[live_callback],
    )
    assert save.call_count != save_weights_only
    assert save_weights.call_count == save_weights_only
    log_artifact.assert_called_with(live_callback.model_file)


@pytest.mark.parametrize("save_weights_only", [True, False])
def test_keras_load_model_on_resume(tmp_dir, xor_model, mocker, save_weights_only):
    import dvclive.keras

    model, x, y = xor_model()

    if save_weights_only:
        model.save_weights("model.h5")
    else:
        model.save("model.h5")

    load_weights = mocker.spy(model, "load_weights")
    load_model = mocker.spy(dvclive.keras, "load_model")

    model.fit(
        x,
        y,
        epochs=1,
        batch_size=1,
        callbacks=[
            DVCLiveCallback(
                model_file="model.h5",
                save_weights_only=save_weights_only,
                resume=True,
            )
        ],
    )

    assert load_model.call_count != save_weights_only
    assert load_weights.call_count == save_weights_only


def test_keras_no_resume_skip_load(tmp_dir, xor_model, mocker):
    model, x, y = xor_model()

    model.save_weights("model.h5")

    load_weights = mocker.spy(model, "load_weights")

    model.fit(
        x,
        y,
        epochs=1,
        batch_size=1,
        callbacks=[
            DVCLiveCallback(
                model_file="model.h5",
                save_weights_only=True,
                resume=False,
            )
        ],
    )

    assert load_weights.call_count == 0


def test_keras_no_existing_model_file_skip_load(tmp_dir, xor_model, mocker):
    model, x, y = xor_model()

    load_weights = mocker.spy(model, "load_weights")

    model.fit(
        x,
        y,
        epochs=1,
        batch_size=1,
        callbacks=[
            DVCLiveCallback(
                model_file="model.h5",
                save_weights_only=True,
                resume=True,
            )
        ],
    )

    assert load_weights.call_count == 0


def test_keras_none_model_file_skip_load(tmp_dir, xor_model, mocker):
    model, x, y = xor_model()

    model.save_weights("model.h5")

    load_weights = mocker.spy(model, "load_weights")

    model.fit(
        x,
        y,
        epochs=1,
        batch_size=1,
        callbacks=[
            DVCLiveCallback(
                save_weights_only=True,
                resume=True,
            )
        ],
    )

    assert load_weights.call_count == 0



from typing import Any


class DvcLiveError(Exception):
    pass


class InvalidDataTypeError(DvcLiveError):
    def __init__(self, name, val):
        self.name = name
        self.val = val
        super().__init__(f"Data '{name}' has not supported type {val}")


class InvalidPlotTypeError(DvcLiveError):
    def __init__(self, name):
        from .plots import SKLEARN_PLOTS

        self.name = name
        super().__init__(
            f"Plot type '{name}' is not supported."
            f"\nSupported types are: {list(SKLEARN_PLOTS)}"
        )


class InvalidParameterTypeError(DvcLiveError):
    def __init__(self, val: Any):
        self.val = val
        super().__init__(f"Parameter type {type(val)} is not supported.")


class InvalidReportModeError(DvcLiveError):
    def __init__(self, val):
        super().__init__(
            f"`report` can only be `None`, `auto`, `html`, `notebook` or `md`. "
            f"Got {val} instead."
        )



import abc
from pathlib import Path


class Data(abc.ABC):
    def __init__(self, name: str, output_folder: str) -> None:
        self.name = name
        self.output_folder: Path = Path(output_folder) / self.subfolder
        self._step: int = -1

    @property
    def step(self) -> int:
        return self._step

    @step.setter
    def step(self, val: int) -> None:
        self._step = val

    @property
    @abc.abstractmethod
    def output_path(self) -> Path:
        pass

    @property
    @abc.abstractmethod
    def subfolder(self):
        pass

    @staticmethod
    @abc.abstractmethod
    def could_log(val) -> bool:
        pass

    @abc.abstractmethod
    def dump(self, val, **kwargs):
        pass



from dvclive.serialize import dump_json

from .custom import CustomPlot


class SKLearnPlot(CustomPlot):
    subfolder = "sklearn"

    @staticmethod
    def could_log(val: object) -> bool:
        if isinstance(val, tuple) and len(val) == 2:  # noqa: PLR2004
            return True
        return False


class Roc(SKLearnPlot):
    def __init__(self, name: str, output_folder: str, **plot_config) -> None:
        plot_config["template"] = plot_config.get("template", "simple")
        plot_config["title"] = plot_config.get(
            "title", "Receiver operating characteristic (ROC)"
        )
        plot_config["x_label"] = plot_config.get("x_label", "False Positive Rate")
        plot_config["y_label"] = plot_config.get("y_label", "True Positive Rate")
        plot_config["x"] = "fpr"
        plot_config["y"] = "tpr"
        super().__init__(name, output_folder, **plot_config)

    def dump(self, val, **kwargs) -> None:
        from sklearn import metrics

        fpr, tpr, roc_thresholds = metrics.roc_curve(
            y_true=val[0], y_score=val[1], **kwargs
        )
        roc = {
            "roc": [
                {"fpr": fp, "tpr": tp, "threshold": t}
                for fp, tp, t in zip(fpr, tpr, roc_thresholds)
            ]
        }
        dump_json(roc, self.output_path)


class PrecisionRecall(SKLearnPlot):
    def __init__(self, name: str, output_folder: str, **plot_config) -> None:
        plot_config["template"] = plot_config.get("template", "simple")
        plot_config["title"] = plot_config.get("title", "Precision-Recall Curve")
        plot_config["x_label"] = plot_config.get("x_label", "Recall")
        plot_config["y_label"] = plot_config.get("y_label", "Precision")
        plot_config["x"] = "recall"
        plot_config["y"] = "precision"
        super().__init__(name, output_folder, **plot_config)

    def dump(self, val, **kwargs) -> None:
        from sklearn import metrics

        precision, recall, prc_thresholds = metrics.precision_recall_curve(
            y_true=val[0], probas_pred=val[1], **kwargs
        )

        prc = {
            "precision_recall": [
                {"precision": p, "recall": r, "threshold": t}
                for p, r, t in zip(precision, recall, prc_thresholds)
            ]
        }
        dump_json(prc, self.output_path)


class Det(SKLearnPlot):
    def __init__(self, name: str, output_folder: str, **plot_config) -> None:
        plot_config["template"] = plot_config.get("template", "simple")
        plot_config["title"] = plot_config.get(
            "title", "Detection error tradeoff (DET)"
        )
        plot_config["x_label"] = plot_config.get("x_label", "False Positive Rate")
        plot_config["y_label"] = plot_config.get("y_label", "False Negative Rate")
        plot_config["x"] = "fpr"
        plot_config["y"] = "fnr"
        super().__init__(name, output_folder, **plot_config)

    def dump(self, val, **kwargs) -> None:
        from sklearn import metrics

        fpr, fnr, roc_thresholds = metrics.det_curve(
            y_true=val[0], y_score=val[1], **kwargs
        )

        det = {
            "det": [
                {"fpr": fp, "fnr": fn, "threshold": t}
                for fp, fn, t in zip(fpr, fnr, roc_thresholds)
            ]
        }
        dump_json(det, self.output_path)


class ConfusionMatrix(SKLearnPlot):
    def __init__(self, name: str, output_folder: str, **plot_config) -> None:
        plot_config["template"] = (
            "confusion_normalized"
            if plot_config.pop("normalized", None)
            else plot_config.get("template", "confusion")
        )
        plot_config["title"] = plot_config.get("title", "Confusion Matrix")
        plot_config["x_label"] = plot_config.get("x_label", "True Label")
        plot_config["y_label"] = plot_config.get("y_label", "Predicted Label")
        plot_config["x"] = "actual"
        plot_config["y"] = "predicted"
        super().__init__(name, output_folder, **plot_config)

    def dump(self, val, **kwargs) -> None:  # noqa: ARG002
        cm = [
            {"actual": str(actual), "predicted": str(predicted)}
            for actual, predicted in zip(val[0], val[1])
        ]
        dump_json(cm, self.output_path)


class Calibration(SKLearnPlot):
    def __init__(self, name: str, output_folder: str, **plot_config) -> None:
        plot_config["template"] = plot_config.get("template", "simple")
        plot_config["title"] = plot_config.get("title", "Calibration Curve")
        plot_config["x_label"] = plot_config.get(
            "x_label", "Mean Predicted Probability"
        )
        plot_config["y_label"] = plot_config.get("y_label", "Fraction of Positives")
        plot_config["x"] = "prob_pred"
        plot_config["y"] = "prob_true"
        super().__init__(name, output_folder, **plot_config)

    def dump(self, val, **kwargs) -> None:
        from sklearn import calibration

        prob_true, prob_pred = calibration.calibration_curve(
            y_true=val[0], y_prob=val[1], **kwargs
        )

        _calibration = {
            "calibration": [
                {"prob_true": pt, "prob_pred": pp}
                for pt, pp in zip(prob_true, prob_pred)
            ]
        }
        dump_json(_calibration, self.output_path)



import os
from sys import platform

import pytest

from dvclive import Live
from dvclive.utils import parse_metrics

try:
    import lightgbm as lgbm
    import numpy as np
    import pandas as pd
    from sklearn import datasets
    from sklearn.model_selection import train_test_split

    from dvclive.lgbm import DVCLiveCallback
except ImportError:
    pytest.skip("skipping lightgbm tests", allow_module_level=True)


@pytest.fixture()
def model_params():
    return {"objective": "multiclass", "n_estimators": 5, "seed": 0}


@pytest.fixture()
def iris_data():
    iris = datasets.load_iris()
    x = pd.DataFrame(iris["data"], columns=iris["feature_names"])
    y = iris["target"]
    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.33, random_state=42
    )
    return (x_train, y_train), (x_test, y_test)


@pytest.mark.skipif(platform == "darwin", reason="LIBOMP Segmentation fault on MacOS")
def test_lgbm_integration(tmp_dir, model_params, iris_data):
    model = lgbm.LGBMClassifier()
    model.set_params(**model_params)

    callback = DVCLiveCallback()
    model.fit(
        iris_data[0][0],
        iris_data[0][1],
        eval_set=(iris_data[1][0], iris_data[1][1]),
        eval_metric=["multi_logloss"],
        callbacks=[callback],
    )

    assert os.path.exists("dvclive")

    logs, _ = parse_metrics(callback.live)
    assert "dvclive/plots/metrics/multi_logloss.tsv" in logs
    assert len(logs) == 1
    assert len(list(logs.values())[0]) == 5


@pytest.mark.skipif(platform == "darwin", reason="LIBOMP Segmentation fault on MacOS")
def test_lgbm_integration_multi_eval(tmp_dir, model_params, iris_data):
    model = lgbm.LGBMClassifier()
    model.set_params(**model_params)

    callback = DVCLiveCallback()
    model.fit(
        iris_data[0][0],
        iris_data[0][1],
        eval_set=[
            (iris_data[0][0], iris_data[0][1]),
            (iris_data[1][0], iris_data[1][1]),
        ],
        eval_metric=["multi_logloss"],
        callbacks=[callback],
    )

    assert os.path.exists("dvclive")

    logs, _ = parse_metrics(callback.live)
    assert "dvclive/plots/metrics/training/multi_logloss.tsv" in logs
    assert "dvclive/plots/metrics/valid_1/multi_logloss.tsv" in logs
    assert len(logs) == 2
    assert len(list(logs.values())[0]) == 5


@pytest.mark.skipif(platform == "darwin", reason="LIBOMP Segmentation fault on MacOS")
def test_lgbm_model_file(tmp_dir, model_params, iris_data):
    model = lgbm.LGBMClassifier()
    model.set_params(**model_params)

    model.fit(
        iris_data[0][0],
        iris_data[0][1],
        eval_set=(iris_data[1][0], iris_data[1][1]),
        eval_metric=["multi_logloss"],
        callbacks=[DVCLiveCallback("lgbm_model")],
    )

    preds = model.predict(iris_data[1][0])
    model2 = lgbm.Booster(model_file="lgbm_model")
    preds2 = model2.predict(iris_data[1][0])
    preds2 = np.argmax(preds2, axis=1)
    assert np.sum(np.abs(preds2 - preds)) == 0


def test_lgbm_pass_logger():
    logger = Live("train_logs")

    assert DVCLiveCallback().live is not logger
    assert DVCLiveCallback(live=logger).live is logger



import pytest

from dvclive.utils import standardize_metric_name


@pytest.mark.parametrize(
    ("framework", "logged", "standardized"),
    [
        ("dvclive.lightning", "epoch", "epoch"),
        ("dvclive.lightning", "train_loss", "train/loss"),
        ("dvclive.lightning", "train_loss_epoch", "train/epoch/loss"),
        ("dvclive.lightning", "train_model_error", "train/model_error"),
        ("dvclive.lightning", "grad_step", "grad_step"),
    ],
)
def test_standardize_metric_name(framework, logged, standardized):
    assert standardize_metric_name(logged, framework) == standardized



# ruff: noqa: ARG002
from typing import Optional

from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)
from transformers.trainer import Trainer

from dvclive import Live
from dvclive.utils import standardize_metric_name


class DVCLiveCallback(TrainerCallback):
    def __init__(self, model_file=None, live: Optional[Live] = None, **kwargs):
        super().__init__()
        self.model_file = model_file
        self.live = live if live is not None else Live(**kwargs)

    def on_log(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        logs = kwargs["logs"]
        for key, value in logs.items():
            self.live.log_metric(standardize_metric_name(key, __name__), value)
        self.live.next_step()

    def on_epoch_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        if self.model_file:
            model = kwargs["model"]
            model.save_pretrained(self.model_file)
            tokenizer = kwargs.get("tokenizer")
            if tokenizer:
                tokenizer.save_pretrained(self.model_file)
            self.live.log_artifact(self.model_file)

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        if args.load_best_model_at_end:
            trainer = Trainer(
                args=args, model=kwargs.get("model"), tokenizer=kwargs.get("tokenizer")
            )
            trainer.save_model()
            self.live.log_artifact(args.output_dir)
        self.live.end()



# ruff: noqa: N806
import os

import pytest

from dvclive import Live
from dvclive.plots import Metric

try:
    import catalyst
    import torch
    from catalyst import dl

    from dvclive.catalyst import DVCLiveCallback
except ImportError:
    pytest.skip("skipping catalyst tests", allow_module_level=True)


@pytest.fixture()
def runner():
    return dl.SupervisedRunner(
        engine=catalyst.utils.torch.get_available_engine(cpu=True),
        input_key="features",
        output_key="logits",
        target_key="targets",
        loss_key="loss",
    )


# see:
# https://github.com/catalyst-team/catalyst/blob/e99f9/tests/catalyst/callbacks/test_batch_overfit.py
@pytest.fixture()
def runner_params():
    from torch.utils.data import DataLoader, TensorDataset

    catalyst.utils.set_global_seed(42)
    num_samples, num_features = int(32e1), int(1e1)
    X, y = torch.rand(num_samples, num_features), torch.rand(num_samples)
    dataset = TensorDataset(X, y)
    loader = DataLoader(dataset, batch_size=32, num_workers=0)
    loaders = {"train": loader, "valid": loader}

    model = torch.nn.Linear(num_features, 1)
    criterion = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters())
    scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, [3, 6])
    return {
        "model": model,
        "criterion": criterion,
        "optimizer": optimizer,
        "scheduler": scheduler,
        "loaders": loaders,
    }


def test_catalyst_callback(tmp_dir, runner, runner_params, mocker):
    callback = DVCLiveCallback()
    live = callback.live
    spy = mocker.spy(live, "end")

    runner.train(
        **runner_params,
        num_epochs=2,
        callbacks=[
            dl.AccuracyCallback(input_key="logits", target_key="targets"),
            callback,
        ],
        logdir="./logs",
        valid_loader="valid",
        valid_metric="loss",
        minimize_valid_metric=True,
        verbose=True,
        load_best_on_end=True,
    )
    spy.assert_called_once()

    assert os.path.exists("dvclive")

    train_path = tmp_dir / "dvclive" / "plots" / Metric.subfolder / "train"
    valid_path = tmp_dir / "dvclive" / "plots" / Metric.subfolder / "valid"

    assert train_path.is_dir()
    assert valid_path.is_dir()
    assert any("accuracy" in x.name for x in train_path.iterdir())
    assert any("accuracy" in x.name for x in valid_path.iterdir())


def test_catalyst_model_file(tmp_dir, runner, runner_params):
    runner.train(
        **runner_params,
        num_epochs=2,
        callbacks=[
            dl.AccuracyCallback(input_key="logits", target_key="targets"),
            DVCLiveCallback("model.pth"),
        ],
        logdir="./logs",
        valid_loader="valid",
        valid_metric="loss",
        minimize_valid_metric=True,
        verbose=True,
        load_best_on_end=True,
    )
    assert (tmp_dir / "model.pth").is_file()


def test_catalyst_pass_logger():
    logger = Live("train_logs")

    assert DVCLiveCallback().live is not logger
    assert DVCLiveCallback(live=logger).live is logger



import json

from dvclive import Live
from dvclive.plots.custom import CustomPlot


def test_log_custom_plot(tmp_dir):
    live = Live()
    out = tmp_dir / live.plots_dir / CustomPlot.subfolder

    datapoints = [{"x": 1, "y": 2}, {"x": 3, "y": 4}]
    live.log_plot(
        "custom_linear",
        datapoints,
        x="x",
        y="y",
        template="linear",
        title="custom_title",
        x_label="x_label",
        y_label="y_label",
    )

    assert json.loads((out / "custom_linear.json").read_text()) == datapoints
    assert live._plots["custom_linear"].plot_config == {
        "template": "linear",
        "title": "custom_title",
        "x": "x",
        "y": "y",
        "x_label": "x_label",
        "y_label": "y_label",
    }



# ruff: noqa: N806
import json

import pytest
from sklearn import calibration, metrics

from dvclive import Live
from dvclive.plots.sklearn import SKLearnPlot


@pytest.fixture()
def y_true_y_pred_y_score():
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import train_test_split

    X, y = make_classification(random_state=0)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)
    clf = RandomForestClassifier(random_state=0)
    clf.fit(X_train, y_train)

    y_pred = clf.predict(X_test)
    y_score = clf.predict_proba(X_test)[:, 1]

    return y_test, y_pred, y_score


def test_log_calibration_curve(tmp_dir, y_true_y_pred_y_score, mocker):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, _, y_score = y_true_y_pred_y_score

    spy = mocker.spy(calibration, "calibration_curve")

    live.log_sklearn_plot("calibration", y_true, y_score)

    spy.assert_called_once_with(y_true, y_score)

    assert (out / "calibration.json").exists()


def test_log_det_curve(tmp_dir, y_true_y_pred_y_score, mocker):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, _, y_score = y_true_y_pred_y_score

    spy = mocker.spy(metrics, "det_curve")

    live.log_sklearn_plot("det", y_true, y_score)

    spy.assert_called_once_with(y_true, y_score)
    assert (out / "det.json").exists()


def test_log_roc_curve(tmp_dir, y_true_y_pred_y_score, mocker):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, _, y_score = y_true_y_pred_y_score

    spy = mocker.spy(metrics, "roc_curve")

    live.log_sklearn_plot("roc", y_true, y_score)

    spy.assert_called_once_with(y_true, y_score)
    assert (out / "roc.json").exists()


def test_log_prc_curve(tmp_dir, y_true_y_pred_y_score, mocker):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, _, y_score = y_true_y_pred_y_score

    spy = mocker.spy(metrics, "precision_recall_curve")

    live.log_sklearn_plot("precision_recall", y_true, y_score)

    spy.assert_called_once_with(y_true, y_score)
    assert (out / "precision_recall.json").exists()


def test_log_confusion_matrix(tmp_dir, y_true_y_pred_y_score, mocker):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, y_pred, _ = y_true_y_pred_y_score

    live.log_sklearn_plot("confusion_matrix", y_true, y_pred)

    cm = json.loads((out / "confusion_matrix.json").read_text())

    assert isinstance(cm, list)
    assert isinstance(cm[0], dict)
    assert cm[0]["actual"] == str(y_true[0])
    assert cm[0]["predicted"] == str(y_pred[0])


def test_dump_kwargs(tmp_dir, y_true_y_pred_y_score, mocker):
    live = Live()

    y_true, _, y_score = y_true_y_pred_y_score

    spy = mocker.spy(metrics, "roc_curve")

    live.log_sklearn_plot("roc", y_true, y_score, drop_intermediate=True)

    spy.assert_called_once_with(y_true, y_score, drop_intermediate=True)


def test_override_on_step(tmp_dir):
    live = Live()

    live.log_sklearn_plot("confusion_matrix", [0, 0], [0, 0])
    live.next_step()
    live.log_sklearn_plot("confusion_matrix", [0, 0], [1, 1])

    plot_path = tmp_dir / live.plots_dir / SKLearnPlot.subfolder
    plot_path = plot_path / "confusion_matrix.json"

    assert json.loads(plot_path.read_text()) == [
        {"actual": "0", "predicted": "1"},
        {"actual": "0", "predicted": "1"},
    ]


def test_cleanup(tmp_dir, y_true_y_pred_y_score):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, y_pred, _ = y_true_y_pred_y_score

    live.log_sklearn_plot("confusion_matrix", y_true, y_pred)

    assert (out / "confusion_matrix.json").exists()

    Live()

    assert not (tmp_dir / live.plots_dir / SKLearnPlot.subfolder).exists()


def test_custom_name(tmp_dir, y_true_y_pred_y_score):
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, y_pred, _ = y_true_y_pred_y_score

    live.log_sklearn_plot("confusion_matrix", y_true, y_pred, name="train/cm")
    live.log_sklearn_plot("confusion_matrix", y_true, y_pred, name="val/cm")
    # ".json" should be stripped from the name
    live.log_sklearn_plot("confusion_matrix", y_true, y_pred, name="cm.json")

    assert (out / "train" / "cm.json").exists()
    assert (out / "val" / "cm.json").exists()
    assert (out / "cm.json").exists()


def test_custom_title(tmp_dir, y_true_y_pred_y_score):
    """https://github.com/iterative/dvclive/issues/453"""
    live = Live()
    out = tmp_dir / live.plots_dir / SKLearnPlot.subfolder

    y_true, y_pred, _ = y_true_y_pred_y_score

    live.log_sklearn_plot(
        "confusion_matrix",
        y_true,
        y_pred,
        name="train/cm",
        title="Train Confusion Matrix",
    )
    live.log_sklearn_plot(
        "confusion_matrix", y_true, y_pred, name="val/cm", title="Val Confusion Matrix"
    )
    assert (out / "train" / "cm.json").exists()
    assert (out / "val" / "cm.json").exists()

    assert live._plots["train/cm"].plot_config["title"] == "Train Confusion Matrix"
    assert live._plots["val/cm"].plot_config["title"] == "Val Confusion Matrix"



from .custom import CustomPlot
from .image import Image
from .metric import Metric
from .sklearn import Calibration, ConfusionMatrix, Det, PrecisionRecall, Roc
from .utils import NumpyEncoder  # noqa: F401

SKLEARN_PLOTS = {
    "calibration": Calibration,
    "confusion_matrix": ConfusionMatrix,
    "det": Det,
    "precision_recall": PrecisionRecall,
    "roc": Roc,
}
PLOT_TYPES = (*SKLEARN_PLOTS.values(), Metric, Image, CustomPlot)



DVCLIVE_LOGLEVEL = "DVCLIVE_LOGLEVEL"
DVCLIVE_OPEN = "DVCLIVE_OPEN"
DVCLIVE_RESUME = "DVCLIVE_RESUME"
DVC_CHECKPOINT = "DVC_CHECKPOINT"
DVC_EXP_BASELINE_REV = "DVC_EXP_BASELINE_REV"
DVC_EXP_NAME = "DVC_EXP_NAME"



import json

NUMPY_INTS = [
    "int_",
    "intc",
    "intp",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
]
NUMPY_FLOATS = ["float_", "float16", "float32", "float64"]
NUMPY_SCALARS = NUMPY_INTS + NUMPY_FLOATS


class NumpyEncoder(json.JSONEncoder):
    def default(self, o):
        if o.__class__.__module__ == "numpy":
            if o.__class__.__name__ in NUMPY_INTS:
                return int(o)
            if o.__class__.__name__ in NUMPY_FLOATS:
                return float(o)
        return super().default(o)



import json
from collections import OrderedDict

from dvclive.error import DvcLiveError


class YAMLError(DvcLiveError):
    pass


class YAMLFileCorruptedError(YAMLError):
    def __init__(self, path):
        super().__init__(path, "YAML file structure is corrupted")


def load_yaml(path, typ="safe"):
    from ruamel.yaml import YAML
    from ruamel.yaml import YAMLError as _YAMLError

    yaml = YAML(typ=typ)
    with open(path, encoding="utf-8") as fd:
        try:
            return yaml.load(fd.read())
        except _YAMLError:
            raise YAMLFileCorruptedError(path) from _YAMLError


def get_yaml():
    from ruamel.yaml import YAML

    yaml = YAML()
    yaml.default_flow_style = False

    # tell Dumper to represent OrderedDict as normal dict
    yaml_repr_cls = yaml.Representer
    yaml_repr_cls.add_representer(OrderedDict, yaml_repr_cls.represent_dict)
    return yaml


def dump_yaml(content, output_file):
    yaml = get_yaml()
    with open(output_file, "w", encoding="utf-8") as fd:
        yaml.dump(content, fd)


def dump_json(content, output_file, indent=4, **kwargs):
    with open(output_file, "w", encoding="utf-8") as f:
        json.dump(content, f, indent=indent, **kwargs)
        f.write("\n")



import os

import pytest

from dvclive import Live
from dvclive.utils import parse_metrics

try:
    import numpy as np
    import pandas as pd
    import xgboost as xgb
    from sklearn import datasets

    from dvclive.xgb import DVCLiveCallback
except ImportError:
    pytest.skip("skipping xgboost tests", allow_module_level=True)


@pytest.fixture()
def train_params():
    return {"objective": "multi:softmax", "num_class": 3, "seed": 0}


@pytest.fixture()
def iris_data():
    iris = datasets.load_iris()
    x = pd.DataFrame(iris["data"], columns=iris["feature_names"])
    y = iris["target"]
    return xgb.DMatrix(x, y)


def test_xgb_integration(tmp_dir, train_params, iris_data, mocker):
    callback = DVCLiveCallback("eval_data")
    live = callback.live
    spy = mocker.spy(live, "end")
    xgb.train(
        train_params,
        iris_data,
        callbacks=[callback],
        num_boost_round=5,
        evals=[(iris_data, "eval_data")],
    )
    spy.assert_called_once()

    assert os.path.exists("dvclive")

    logs, _ = parse_metrics(callback.live)
    assert len(logs) == 1
    assert len(list(logs.values())[0]) == 5


def test_xgb_model_file(tmp_dir, train_params, iris_data):
    model = xgb.train(
        train_params,
        iris_data,
        callbacks=[DVCLiveCallback("eval_data", model_file="model_xgb.json")],
        num_boost_round=5,
        evals=[(iris_data, "eval_data")],
    )

    preds = model.predict(iris_data)
    model2 = xgb.Booster(model_file="model_xgb.json")
    preds2 = model2.predict(iris_data)
    assert np.sum(np.abs(preds2 - preds)) == 0


def test_xgb_pass_logger():
    logger = Live("train_logs")

    assert DVCLiveCallback("eval_data").live is not logger
    assert DVCLiveCallback("eval_data", live=logger).live is logger



import shutil
from pathlib import Path

from dvclive import Live
from dvclive.serialize import load_yaml


def test_log_artifact(tmp_dir, dvc_repo):
    data = tmp_dir / "data"
    data.touch()
    with Live() as live:
        live.log_artifact("data")
    assert data.with_suffix(".dvc").exists()


def test_log_artifact_on_existing_dvc_file(tmp_dir, dvc_repo):
    data = tmp_dir / "data"
    data.write_text("foo")
    with Live() as live:
        live.log_artifact("data")

    prev_content = data.with_suffix(".dvc").read_text()

    with Live() as live:
        data.write_text("bar")
        live.log_artifact("data")

    assert data.with_suffix(".dvc").read_text() != prev_content


def test_log_artifact_twice(tmp_dir, dvc_repo):
    data = tmp_dir / "data"
    with Live() as live:
        for i in range(2):
            data.write_text(str(i))
            live.log_artifact("data")
    assert data.with_suffix(".dvc").exists()


def test_log_artifact_with_save_dvc_exp(tmp_dir, mocker, mocked_dvc_repo):
    stage = mocker.MagicMock()
    stage.addressing = "data"
    mocked_dvc_repo.add.return_value = [stage]
    with Live(save_dvc_exp=True) as live:
        live.log_artifact("data")
    mocked_dvc_repo.experiments.save.assert_called_with(
        name=live._exp_name,
        include_untracked=[live.dir, "data", ".gitignore"],
        force=True,
        message=None,
    )


def test_log_artifact_type_model(tmp_dir, mocked_dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        live.log_artifact("model.pth", type="model")

    assert load_yaml(live.dvc_file) == {
        "artifacts": {"model": {"path": "../model.pth", "type": "model"}}
    }


def test_log_artifact_dvc_symlink(tmp_dir, dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        live._dvc_repo.cache.local.cache_types = ["symlink"]
        live.log_artifact("model.pth", type="model")

    assert load_yaml(live.dvc_file) == {
        "artifacts": {"model": {"path": "../model.pth", "type": "model"}}
    }


def test_log_artifact_copy(tmp_dir, dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        live.log_artifact("model.pth", copy=True)

    artifacts_dir = Path(live.artifacts_dir)
    assert (artifacts_dir / "model.pth").exists()
    assert (artifacts_dir / "model.pth.dvc").exists()

    assert load_yaml(live.dvc_file) == {
        "artifacts": {"model": {"path": "artifacts/model.pth"}}
    }


def test_log_artifact_copy_overwrite(tmp_dir, dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        artifacts_dir = Path(live.artifacts_dir)
        # testing with symlink cache to make sure that DVC protected mode
        # does not prevent the overwrite
        live._dvc_repo.cache.local.cache_types = ["symlink"]
        live.log_artifact("model.pth", copy=True)
        assert (artifacts_dir / "model.pth").is_symlink()
        live.log_artifact("model.pth", copy=True)

    assert (artifacts_dir / "model.pth").exists()
    assert (artifacts_dir / "model.pth.dvc").exists()

    assert load_yaml(live.dvc_file) == {
        "artifacts": {"model": {"path": "artifacts/model.pth"}}
    }


def test_log_artifact_copy_directory_overwrite(tmp_dir, dvc_repo):
    model_path = Path(tmp_dir / "weights")
    model_path.mkdir()
    (tmp_dir / "weights" / "model-epoch-1.pth").touch()

    with Live() as live:
        artifacts_dir = Path(live.artifacts_dir)
        # testing with symlink cache to make sure that DVC protected mode
        # does not prevent the overwrite
        live._dvc_repo.cache.local.cache_types = ["symlink"]
        live.log_artifact(model_path, copy=True)
        assert (artifacts_dir / "weights" / "model-epoch-1.pth").is_symlink()

        shutil.rmtree(model_path)
        model_path.mkdir()
        (tmp_dir / "weights" / "model-epoch-10.pth").write_text("Model weights")
        (tmp_dir / "weights" / "best.pth").write_text("Best model weights")
        live.log_artifact(model_path, copy=True)

    assert (artifacts_dir / "weights").exists()
    assert (artifacts_dir / "weights" / "best.pth").is_symlink()
    assert (artifacts_dir / "weights" / "best.pth").read_text() == "Best model weights"
    assert (artifacts_dir / "weights" / "model-epoch-10.pth").is_symlink()
    assert len(list((artifacts_dir / "weights").iterdir())) == 2

    assert load_yaml(live.dvc_file) == {
        "artifacts": {"weights": {"path": "artifacts/weights"}}
    }


def test_log_artifact_type_model_provided_name(tmp_dir, mocked_dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        live.log_artifact("model.pth", type="model", name="custom")

    assert load_yaml(live.dvc_file) == {
        "artifacts": {"custom": {"path": "../model.pth", "type": "model"}}
    }


def test_log_artifact_type_model_on_step_and_final(tmp_dir, mocked_dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        for _ in range(3):
            live.log_artifact("model.pth", type="model")
            live.next_step()
        live.log_artifact("model.pth", type="model", labels=["final"])
    assert load_yaml(live.dvc_file) == {
        "artifacts": {
            "model": {"path": "../model.pth", "type": "model", "labels": ["final"]},
        },
        "metrics": ["metrics.json"],
    }


def test_log_artifact_type_model_on_step(tmp_dir, mocked_dvc_repo):
    (tmp_dir / "model.pth").touch()

    with Live() as live:
        for _ in range(3):
            live.log_artifact("model.pth", type="model")
            live.next_step()

    assert load_yaml(live.dvc_file) == {
        "artifacts": {
            "model": {"path": "../model.pth", "type": "model"},
        },
        "metrics": ["metrics.json"],
    }


def test_log_artifact_attrs(tmp_dir, mocked_dvc_repo):
    (tmp_dir / "model.pth").touch()

    attrs = {
        "type": "model",
        "name": "foo",
        "desc": "bar",
        "labels": ["foo"],
        "meta": {"foo": "bar"},
    }
    with Live() as live:
        live.log_artifact("model.pth", **attrs)
    attrs.pop("name")
    assert load_yaml(live.dvc_file) == {
        "artifacts": {
            "foo": {"path": "../model.pth", **attrs},
        }
    }



from typing import Optional

from dvclive import Live


class DVCLiveCallback:
    def __init__(self, model_file=None, live: Optional[Live] = None, **kwargs):
        super().__init__()
        self.model_file = model_file
        self.live = live if live is not None else Live(**kwargs)

    def __call__(self, env):
        multi_eval = len(env.evaluation_result_list) > 1
        for eval_result in env.evaluation_result_list:
            data_name, eval_name, result = eval_result[:3]
            self.live.log_metric(
                f"{data_name}/{eval_name}" if multi_eval else eval_name, result
            )

        if self.model_file:
            env.model.save_model(self.model_file)
        self.live.next_step()



# ruff: noqa: ARG002
from dvclive import Live


class DVCLiveCallback:
    def __init__(self, metric_name="metric", **kwargs) -> None:
        kwargs["dir"] = kwargs.get("dir", "dvclive-optuna")
        kwargs.pop("save_dvc_exp", None)
        self.metric_name = metric_name
        self.live_kwargs = kwargs

    def __call__(self, study, trial) -> None:
        with Live(save_dvc_exp=True, **self.live_kwargs) as live:
            self._log_metrics(trial.values, live)
            live.log_params(trial.params)

    def _log_metrics(self, values, live):
        if values is None:
            return

        if isinstance(self.metric_name, str):
            if len(values) > 1:
                # Broadcast default name for multi-objective optimization.
                names = [f"{self.metric_name}_{i}" for i in range(len(values))]

            else:
                names = [self.metric_name]

        elif len(self.metric_name) != len(values):
            raise ValueError(
                "Running multi-objective optimization "
                "with {} objective values, but {} names specified. "
                "Match objective values and names,"
                "or use default broadcasting.".format(
                    len(values), len(self.metric_name)
                )
            )

        else:
            names = [*self.metric_name]

        metrics = dict(zip(names, values))
        for k, v in metrics.items():
            live.summary[k] = v



from pathlib import Path

import pytest
from dvc_studio_client.env import DVC_STUDIO_REPO_URL, DVC_STUDIO_TOKEN
from dvc_studio_client.post_live_metrics import STUDIO_URL
from PIL import Image as ImagePIL

from dvclive import Live
from dvclive.env import DVC_EXP_BASELINE_REV, DVC_EXP_NAME
from dvclive.plots import Image, Metric
from dvclive.studio import _adapt_image, get_dvc_studio_config


def test_post_to_studio(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    live = Live(save_dvc_exp=True)
    live.log_param("fooparam", 1)

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    params_path = Path(live.params_file).as_posix()
    foo_path = (Path(live.plots_dir) / Metric.subfolder / "foo.tsv").as_posix()

    mocked_post, _ = mocked_studio_post

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "start",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )

    live.log_metric("foo", 1)

    live.next_step()
    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "step": 0,
            "metrics": {metrics_path: {"data": {"step": 0, "foo": 1}}},
            "params": {params_path: {"fooparam": 1}},
            "plots": {f"{dvc_path}::{foo_path}": {"data": [{"step": 0, "foo": 1.0}]}},
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )

    live.log_metric("foo", 2)

    live.next_step()
    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "step": 1,
            "metrics": {metrics_path: {"data": {"step": 1, "foo": 2}}},
            "params": {params_path: {"fooparam": 1}},
            "plots": {f"{dvc_path}::{foo_path}": {"data": [{"step": 1, "foo": 2.0}]}},
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )

    live.end()
    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "done",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "experiment_rev": live._experiment_rev,
            "name": live._exp_name,
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


def test_post_to_studio_failed_data_request(
    tmp_dir, mocker, mocked_dvc_repo, mocked_studio_post
):
    mocked_post, valid_response = mocked_studio_post

    live = Live(save_dvc_exp=True)

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    foo_path = (Path(live.plots_dir) / Metric.subfolder / "foo.tsv").as_posix()

    error_response = mocker.MagicMock()
    error_response.status_code = 400
    mocker.patch("requests.post", return_value=error_response)
    live.log_metric("foo", 1)
    live.next_step()

    mocked_post = mocker.patch("requests.post", return_value=valid_response)
    live.log_metric("foo", 2)
    live.next_step()
    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "step": 1,
            "metrics": {metrics_path: {"data": {"step": 1, "foo": 2}}},
            "plots": {
                f"{dvc_path}::{foo_path}": {
                    "data": [
                        {"step": 0, "foo": 1.0},
                        {"step": 1, "foo": 2.0},
                    ]
                }
            },
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


def test_post_to_studio_failed_start_request(
    tmp_dir, mocker, mocked_dvc_repo, mocked_studio_post
):
    mocked_response = mocker.MagicMock()
    mocked_response.status_code = 400
    mocked_post = mocker.patch("requests.post", return_value=mocked_response)

    live = Live(save_dvc_exp=True)

    live.log_metric("foo", 1)
    live.next_step()

    live.log_metric("foo", 2)
    live.next_step()

    assert mocked_post.call_count == 1


def test_post_to_studio_end_only_once(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    mocked_post, _ = mocked_studio_post
    with Live(save_dvc_exp=True) as live:
        live.log_metric("foo", 1)
        live.next_step()

    assert mocked_post.call_count == 3
    live.end()
    assert mocked_post.call_count == 3


@pytest.mark.studio()
def test_post_to_studio_skip_on_env_var(
    tmp_dir, mocked_dvc_repo, mocked_studio_post, monkeypatch
):
    mocked_post, _ = mocked_studio_post

    monkeypatch.setenv(DVC_EXP_BASELINE_REV, "f" * 40)
    monkeypatch.setenv(DVC_EXP_NAME, "bar")

    with Live() as live:
        live.log_metric("foo", 1)

    assert mocked_post.call_count == 1


@pytest.mark.studio()
def test_post_to_studio_dvc_studio_config(
    tmp_dir, mocker, mocked_dvc_repo, mocked_studio_post, monkeypatch
):
    mocked_post, _ = mocked_studio_post

    monkeypatch.setenv(DVC_EXP_BASELINE_REV, "f" * 40)
    monkeypatch.setenv(DVC_EXP_NAME, "bar")

    mocked_dvc_repo.config = {"studio": {"token": "token"}}

    with Live() as live:
        live.log_metric("foo", 1)

    assert mocked_post.call_count == 1


@pytest.mark.studio()
def test_post_to_studio_skip_if_no_token(
    tmp_dir,
    mocker,
    monkeypatch,
    mocked_dvc_repo,
):
    mocked_post = mocker.patch("dvclive.live.post_live_metrics", return_value=None)

    monkeypatch.setenv(DVC_EXP_BASELINE_REV, "f" * 40)
    monkeypatch.setenv(DVC_EXP_NAME, "bar")

    mocked_dvc_repo.config = {}

    with Live() as live:
        live.log_metric("foo", 1)
        live.next_step()

    assert mocked_post.call_count == 0


def test_post_to_studio_include_prefix_if_needed(
    tmp_dir, mocked_dvc_repo, mocked_studio_post
):
    mocked_post, _ = mocked_studio_post
    # Create dvclive/dvc.yaml
    live = Live("custom_dir", save_dvc_exp=True)
    live.log_metric("foo", 1)
    live.next_step()

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    foo_path = (Path(live.plots_dir) / Metric.subfolder / "foo.tsv").as_posix()

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "step": 0,
            "metrics": {metrics_path: {"data": {"step": 0, "foo": 1}}},
            "plots": {f"{dvc_path}::{foo_path}": {"data": [{"step": 0, "foo": 1.0}]}},
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


def test_post_to_studio_shorten_names(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    mocked_post, _ = mocked_studio_post

    live = Live(save_dvc_exp=True)
    live.log_metric("eval/loss", 1)
    live.next_step()

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    plots_path = Path(live.plots_dir)
    loss_path = (plots_path / Metric.subfolder / "eval/loss.tsv").as_posix()

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "step": 0,
            "metrics": {metrics_path: {"data": {"step": 0, "eval": {"loss": 1}}}},
            "plots": {f"{dvc_path}::{loss_path}": {"data": [{"step": 0, "loss": 1.0}]}},
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


@pytest.mark.studio()
def test_post_to_studio_inside_dvc_exp(
    tmp_dir, mocker, monkeypatch, mocked_studio_post
):
    mocked_post, _ = mocked_studio_post
    mocker.patch("dvclive.live.get_dvc_repo", return_value=None)

    monkeypatch.setenv(DVC_EXP_BASELINE_REV, "f" * 40)
    monkeypatch.setenv(DVC_EXP_NAME, "bar")

    with Live() as live:
        live.log_metric("foo", 1)

    assert mocked_post.call_count == 1


@pytest.mark.studio()
def test_post_to_studio_inside_subdir(
    tmp_dir, dvc_repo, mocker, monkeypatch, mocked_studio_post
):
    mocked_post, _ = mocked_studio_post
    subdir = tmp_dir / "subdir"
    subdir.mkdir()
    monkeypatch.chdir(subdir)

    live = Live(save_dvc_exp=True)
    live.log_metric("foo", 1)
    live.next_step()

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    foo_path = (Path(live.plots_dir) / Metric.subfolder / "foo.tsv").as_posix()

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": live._baseline_rev,
            "name": live._exp_name,
            "step": 0,
            "metrics": {f"subdir/{metrics_path}": {"data": {"step": 0, "foo": 1}}},
            "plots": {
                f"subdir/{dvc_path}::subdir/{foo_path}": {
                    "data": [{"step": 0, "foo": 1.0}]
                }
            },
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


@pytest.mark.studio()
def test_post_to_studio_inside_subdir_dvc_exp(
    tmp_dir, dvc_repo, monkeypatch, mocked_studio_post
):
    mocked_post, _ = mocked_studio_post
    subdir = tmp_dir / "subdir"
    subdir.mkdir()
    monkeypatch.chdir(subdir)

    monkeypatch.setenv(DVC_EXP_BASELINE_REV, "f" * 40)
    monkeypatch.setenv(DVC_EXP_NAME, "bar")

    live = Live()
    live.log_metric("foo", 1)
    live.next_step()

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    foo_path = (Path(live.plots_dir) / Metric.subfolder / "foo.tsv").as_posix()

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": live._baseline_rev,
            "name": live._exp_name,
            "step": 0,
            "metrics": {f"subdir/{metrics_path}": {"data": {"step": 0, "foo": 1}}},
            "plots": {
                f"subdir/{dvc_path}::subdir/{foo_path}": {
                    "data": [{"step": 0, "foo": 1.0}]
                }
            },
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


def test_post_to_studio_requires_exp(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    assert Live()._studio_events_to_skip == {"start", "data", "done"}
    assert not Live(save_dvc_exp=True)._studio_events_to_skip


def test_get_dvc_studio_config_none(mocker):
    mocker.patch("dvclive.live.get_dvc_repo", return_value=None)
    live = Live()
    assert get_dvc_studio_config(live) == {}


def test_get_dvc_studio_config_env_var(monkeypatch, mocker):
    monkeypatch.setenv(DVC_STUDIO_TOKEN, "token")
    monkeypatch.setenv(DVC_STUDIO_REPO_URL, "repo_url")
    mocker.patch("dvclive.live.get_dvc_repo", return_value=None)
    live = Live()
    assert get_dvc_studio_config(live) == {
        "token": "token",
        "repo_url": "repo_url",
        "url": STUDIO_URL,
    }


def test_get_dvc_studio_config_dvc_repo(mocked_dvc_repo):
    mocked_dvc_repo.config = {"studio": {"token": "token", "repo_url": "repo_url"}}
    live = Live()
    assert get_dvc_studio_config(live) == {
        "token": "token",
        "repo_url": "repo_url",
        "url": STUDIO_URL,
    }


def test_post_to_studio_images(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    mocked_post, _ = mocked_studio_post

    live = Live(save_dvc_exp=True)
    live.log_image("foo.png", ImagePIL.new("RGB", (10, 10), (0, 0, 0)))
    live.next_step()

    dvc_path = Path(live.dvc_file).as_posix()
    metrics_path = Path(live.metrics_file).as_posix()
    foo_path = (Path(live.plots_dir) / Image.subfolder / "foo.png").as_posix()

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "data",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": live._baseline_rev,
            "name": live._exp_name,
            "step": 0,
            "metrics": {f"{metrics_path}": {"data": {"step": 0}}},
            "plots": {
                f"{dvc_path}::{foo_path}": {"image": _adapt_image(foo_path)},
            },
            "client": "dvclive",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )


def test_post_to_studio_message(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    live = Live(save_dvc_exp=True, exp_message="Custom message")

    mocked_post, _ = mocked_studio_post

    mocked_post.assert_called_with(
        "https://0.0.0.0/api/live",
        json={
            "type": "start",
            "repo_url": "STUDIO_REPO_URL",
            "baseline_sha": "f" * 40,
            "name": live._exp_name,
            "client": "dvclive",
            "message": "Custom message",
        },
        headers={
            "Authorization": "token STUDIO_TOKEN",
            "Content-type": "application/json",
        },
        timeout=(30, 5),
    )



import pytest

from dvclive.serialize import load_yaml
from dvclive.utils import parse_json

try:
    import optuna

    from dvclive.optuna import DVCLiveCallback
except ImportError:
    pytest.skip("skipping optuna tests", allow_module_level=True)


def objective(trial):
    x = trial.suggest_float("x", -10, 10)
    return (x - 2) ** 2


def test_optuna_(tmp_dir, mocked_dvc_repo):
    n_trials = 5
    metric_name = "custom_name"
    callback = DVCLiveCallback(metric_name=metric_name)
    study = optuna.create_study()

    study.optimize(objective, n_trials=n_trials, callbacks=[callback])

    assert mocked_dvc_repo.experiments.save.call_count == n_trials

    metrics = parse_json("dvclive-optuna/metrics.json")
    assert metric_name in metrics
    params = load_yaml("dvclive-optuna/params.yaml")
    assert "x" in params

    assert not (tmp_dir / "dvclive-optuna" / "plots").exists()



# ruff: noqa: SLF001
import base64
import os
from pathlib import Path

from dvc_studio_client.post_live_metrics import get_studio_config

from dvclive.serialize import load_yaml
from dvclive.utils import parse_metrics


def _get_unsent_datapoints(plot, latest_step):
    return [x for x in plot if int(x["step"]) > latest_step]


def _cast_to_numbers(datapoints):
    for datapoint in datapoints:
        for k, v in datapoint.items():
            if k == "step":
                datapoint[k] = int(v)
            elif k == "timestamp":
                continue
            else:
                datapoint[k] = float(v)
    return datapoints


def _rel_path(path, dvc_root_path):
    absolute_path = Path(path).resolve()
    return str(absolute_path.relative_to(dvc_root_path).as_posix())


def _adapt_plot_name(live, name):
    if live._dvc_repo is not None:
        name = _rel_path(name, live._dvc_repo.root_dir)
    if os.path.isfile(live.dvc_file):
        dvc_file = live.dvc_file
        if live._dvc_repo is not None:
            dvc_file = _rel_path(live.dvc_file, live._dvc_repo.root_dir)
        name = f"{dvc_file}::{name}"
    return name


def _adapt_plot_datapoints(live, plot):
    datapoints = _get_unsent_datapoints(plot, live._latest_studio_step)
    return _cast_to_numbers(datapoints)


def _adapt_image(image_path):
    with open(image_path, "rb") as fobj:
        return base64.b64encode(fobj.read()).decode("utf-8")


def _adapt_images(live):
    return {
        _adapt_plot_name(live, image.output_path): {
            "image": _adapt_image(image.output_path)
        }
        for image in live._images.values()
        if image.step > live._latest_studio_step
    }


def get_studio_updates(live):
    if os.path.isfile(live.params_file):
        params_file = live.params_file
        if live._dvc_repo is not None:
            params_file = _rel_path(params_file, live._dvc_repo.root_dir)
        params = {params_file: load_yaml(live.params_file)}
    else:
        params = {}

    plots, metrics = parse_metrics(live)

    metrics_file = live.metrics_file
    if live._dvc_repo is not None:
        metrics_file = _rel_path(metrics_file, live._dvc_repo.root_dir)
    metrics = {metrics_file: {"data": metrics}}

    plots = {
        _adapt_plot_name(live, name): _adapt_plot_datapoints(live, plot)
        for name, plot in plots.items()
    }
    plots = {k: {"data": v} for k, v in plots.items()}

    plots.update(_adapt_images(live))

    return metrics, params, plots


def get_dvc_studio_config(live):
    config = {}
    if live._dvc_repo:
        config = live._dvc_repo.config.get("studio")
    return get_studio_config(dvc_studio_config=config)



import numpy as np
import pytest
from PIL import Image

from dvclive import Live
from dvclive.plots import Image as LiveImage


# From https://stackoverflow.com/questions/5165317/how-can-i-extend-image-class
class ExtendedImage(Image.Image):
    def __init__(self, img):
        self._img = img

    def __getattr__(self, key):
        return getattr(self._img, key)


def test_pil(tmp_dir):
    live = Live()
    img = Image.new("RGB", (10, 10), (250, 250, 250))
    live.log_image("image.png", img)

    assert (tmp_dir / live.plots_dir / LiveImage.subfolder / "image.png").exists()


def test_invalid_extension(tmp_dir):
    live = Live()
    img = Image.new("RGB", (10, 10), (250, 250, 250))
    with pytest.raises(ValueError, match="unknown file extension"):
        live.log_image("image.foo", img)


@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 3), (10, 10, 4)])
def test_numpy(tmp_dir, shape):
    from PIL import Image as ImagePIL

    live = Live()
    img = np.ones(shape, np.uint8) * 255
    live.log_image("image.png", img)

    img_path = tmp_dir / live.plots_dir / LiveImage.subfolder / "image.png"
    assert img_path.exists()

    val = np.asarray(ImagePIL.open(img_path))
    assert np.array_equal(val, img)


def test_path(tmp_dir):
    import numpy as np
    from PIL import Image as ImagePIL

    live = Live()
    image_data = np.random.randint(0, 255, (100, 100, 3)).astype(np.uint8)
    pil_image = ImagePIL.fromarray(image_data)
    image_path = tmp_dir / "temp.png"
    pil_image.save(image_path)

    live = Live()
    live.log_image("foo.png", image_path)
    live.end()

    plot_file = tmp_dir / live.plots_dir / "images" / "foo.png"
    assert plot_file.exists()

    val = np.asarray(ImagePIL.open(plot_file))
    assert np.array_equal(val, image_data)


def test_override_on_step(tmp_dir):
    live = Live()

    zeros = np.zeros((2, 2, 3), np.uint8)
    live.log_image("image.png", zeros)

    live.next_step()

    ones = np.ones((2, 2, 3), np.uint8)
    live.log_image("image.png", ones)

    img_path = tmp_dir / live.plots_dir / LiveImage.subfolder / "image.png"
    assert np.array_equal(np.array(Image.open(img_path)), ones)


def test_cleanup(tmp_dir):
    live = Live()
    img = np.ones((10, 10, 3), np.uint8)
    live.log_image("image.png", img)

    assert (tmp_dir / live.plots_dir / LiveImage.subfolder / "image.png").exists()

    Live()

    assert not (tmp_dir / live.plots_dir / LiveImage.subfolder).exists()


def test_custom_class(tmp_dir):
    live = Live()
    img = Image.new("RGB", (10, 10), (250, 250, 250))
    extended_img = ExtendedImage(img)
    live.log_image("image.png", extended_img)

    assert (tmp_dir / live.plots_dir / LiveImage.subfolder / "image.png").exists()



from pathlib import Path, PurePath

from .base import Data


class Image(Data):
    suffixes = [".jpg", ".jpeg", ".gif", ".png"]
    subfolder = "images"

    @property
    def output_path(self) -> Path:
        _path = self.output_folder / self.name
        _path.parent.mkdir(exist_ok=True, parents=True)
        return _path

    @staticmethod
    def could_log(val: object) -> bool:
        acceptable = {
            ("numpy", "ndarray"),
            ("PIL.Image", "Image"),
        }
        for cls in type(val).mro():
            if (cls.__module__, cls.__name__) in acceptable:
                return True
        if isinstance(val, (PurePath, str)):
            return True
        return False

    def dump(self, val, **kwargs) -> None:  # noqa: ARG002
        if val.__class__.__module__ == "numpy":
            from PIL import Image as ImagePIL

            pil_image = ImagePIL.fromarray(val)
        else:
            pil_image = val
        pil_image.save(self.output_path)



import json
import logging
import os

import pytest

from dvclive import Live, env
from dvclive.error import InvalidDataTypeError, InvalidParameterTypeError
from dvclive.plots import Metric
from dvclive.serialize import load_yaml
from dvclive.utils import parse_metrics, parse_tsv


def read_history(live, metric):
    history, _ = parse_metrics(live)
    steps = []
    values = []
    name = os.path.join(live.plots_dir, Metric.subfolder, f"{metric}.tsv")
    for e in history[name]:
        steps.append(int(e["step"]))
        values.append(float(e[metric]))
    return steps, values


def read_latest(live, metric_name):
    _, latest = parse_metrics(live)
    return latest["step"], latest[metric_name]


def test_logging_no_step(tmp_dir):
    dvclive = Live("logs")

    dvclive.log_metric("m1", 1, plot=False)
    dvclive.make_summary()

    assert not (tmp_dir / "logs" / "plots" / "metrics" / "m1.tsv").is_file()
    assert (tmp_dir / dvclive.metrics_file).is_file()

    s = load_yaml(dvclive.metrics_file)
    assert s["m1"] == 1
    assert "step" not in s


@pytest.mark.parametrize(
    ("param_name", "param_value"),
    [
        ("param_string", "value"),
        ("param_int", 42),
        ("param_float", 42.0),
        ("param_bool_true", True),
        ("param_bool_false", False),
        ("param_list", [1, 2, 3]),
        (
            "param_dict_simple",
            {"str": "value", "int": 42, "bool": True, "list": [1, 2, 3]},
        ),
        (
            "param_dict_nested",
            {
                "str": "value",
                "int": 42,
                "bool": True,
                "list": [1, 2, 3],
                "dict": {"nested-str": "value", "nested-int": 42},
            },
        ),
    ],
)
def test_log_param(tmp_dir, param_name, param_value):
    dvclive = Live()

    dvclive.log_param(param_name, param_value)

    s = load_yaml(dvclive.params_file)
    assert s[param_name] == param_value


def test_log_params(tmp_dir):
    dvclive = Live()
    params = {
        "param_string": "value",
        "param_int": 42,
        "param_float": 42.0,
        "param_bool_true": True,
        "param_bool_false": False,
    }

    dvclive.log_params(params)

    s = load_yaml(dvclive.params_file)
    assert s == params


@pytest.mark.parametrize("resume", [False, True])
def test_log_params_resume(tmp_dir, resume):
    dvclive = Live(resume=resume)
    dvclive.log_param("param", 42)

    dvclive = Live(resume=resume)
    assert ("param" in dvclive._params) == resume


def test_log_param_custom_obj(tmp_dir):
    dvclive = Live("logs")

    class Dummy:
        val = 42

    param_value = Dummy()

    with pytest.raises(InvalidParameterTypeError):
        dvclive.log_param("param_complex", param_value)


@pytest.mark.parametrize("path", ["logs", os.path.join("subdir", "logs")])
def test_logging_step(tmp_dir, path):
    dvclive = Live(path)
    dvclive.log_metric("m1", 1)
    dvclive.next_step()
    assert (tmp_dir / dvclive.dir).is_dir()
    assert (tmp_dir / dvclive.plots_dir / Metric.subfolder / "m1.tsv").is_file()
    assert (tmp_dir / dvclive.metrics_file).is_file()

    s = load_yaml(dvclive.metrics_file)
    assert s["m1"] == 1
    assert s["step"] == 0


def test_nested_logging(tmp_dir):
    dvclive = Live("logs")

    out = tmp_dir / dvclive.plots_dir / Metric.subfolder

    dvclive.log_metric("train/m1", 1)
    dvclive.log_metric("val/val_1/m1", 1)
    dvclive.log_metric("val/val_1/m2", 1)

    dvclive.next_step()

    assert (out / "val" / "val_1").is_dir()
    assert (out / "train" / "m1.tsv").is_file()
    assert (out / "val" / "val_1" / "m1.tsv").is_file()
    assert (out / "val" / "val_1" / "m2.tsv").is_file()

    assert "m1" in parse_tsv(out / "train" / "m1.tsv")[0]
    assert "m1" in parse_tsv(out / "val" / "val_1" / "m1.tsv")[0]
    assert "m2" in parse_tsv(out / "val" / "val_1" / "m2.tsv")[0]

    summary = load_yaml(dvclive.metrics_file)

    assert summary["train"]["m1"] == 1
    assert summary["val"]["val_1"]["m1"] == 1
    assert summary["val"]["val_1"]["m2"] == 1


@pytest.mark.parametrize(
    "html",
    [True, False],
)
def test_cleanup(tmp_dir, html):
    dvclive = Live("logs", report="html" if html else None)
    dvclive.log_metric("m1", 1)
    dvclive.next_step()

    html_path = tmp_dir / dvclive.dir / "report.html"
    if html:
        html_path.touch()

    (tmp_dir / "logs" / "some_user_file.txt").touch()

    assert (tmp_dir / dvclive.plots_dir / Metric.subfolder / "m1.tsv").is_file()
    assert (tmp_dir / dvclive.metrics_file).is_file()
    assert html_path.is_file() == html

    dvclive = Live("logs")

    assert (tmp_dir / "logs" / "some_user_file.txt").is_file()
    assert not (tmp_dir / dvclive.plots_dir / Metric.subfolder).exists()
    assert not (tmp_dir / dvclive.metrics_file).is_file()
    assert not (html_path).is_file()


def test_cleanup_params(tmp_dir):
    dvclive = Live("logs")
    dvclive.log_param("param", 42)

    assert os.path.isfile(dvclive.params_file)

    dvclive = Live("logs")
    assert not os.path.exists(dvclive.params_file)


@pytest.mark.parametrize(
    ("resume", "steps", "metrics"),
    [(True, [0, 1, 2, 3], [0.9, 0.8, 0.7, 0.6]), (False, [0, 1], [0.7, 0.6])],
)
def test_continue(tmp_dir, resume, steps, metrics):
    dvclive = Live("logs")

    for metric in [0.9, 0.8]:
        dvclive.log_metric("metric", metric)
        dvclive.next_step()

    assert read_history(dvclive, "metric") == ([0, 1], [0.9, 0.8])
    assert read_latest(dvclive, "metric") == (1, 0.8)

    dvclive = Live("logs", resume=resume)

    for new_metric in [0.7, 0.6]:
        dvclive.log_metric("metric", new_metric)
        dvclive.next_step()

    assert read_history(dvclive, "metric") == (steps, metrics)
    assert read_latest(dvclive, "metric") == (steps[-1], metrics[-1])


def test_resume_on_first_init(tmp_dir):
    dvclive = Live(resume=True)

    assert dvclive._step == 0


def test_resume_env_var(tmp_dir, monkeypatch):
    assert not Live()._resume

    monkeypatch.setenv(env.DVCLIVE_RESUME, True)
    assert Live()._resume


@pytest.mark.parametrize("metric", ["m1", os.path.join("train", "m1")])
def test_allow_step_override(tmp_dir, metric):
    dvclive = Live("logs")

    dvclive.log_metric(metric, 1.0)
    dvclive.log_metric(metric, 2.0)


def test_custom_steps(tmp_dir):
    dvclive = Live("logs")

    steps = [0, 62, 1000]
    metrics = [0.9, 0.8, 0.7]

    for step, metric in zip(steps, metrics):
        dvclive.step = step
        dvclive.log_metric("m", metric)
        dvclive.make_summary()

    assert read_history(dvclive, "m") == (steps, metrics)
    assert read_latest(dvclive, "m") == (steps[-1], metrics[-1])


def test_log_reset_with_set_step(tmp_dir):
    dvclive = Live()

    for i in range(3):
        dvclive.step = i
        dvclive.log_metric("train_m", 1)
        dvclive.make_summary()

    for i in range(3):
        dvclive.step = i
        dvclive.log_metric("val_m", 1)
        dvclive.make_summary()

    assert read_history(dvclive, "train_m") == ([0, 1, 2], [1, 1, 1])
    assert read_history(dvclive, "val_m") == ([0, 1, 2], [1, 1, 1])
    assert read_latest(dvclive, "train_m") == (2, 1)
    assert read_latest(dvclive, "val_m") == (2, 1)


@pytest.mark.parametrize("invalid_type", [{0: 1}, [0, 1], "foo", (0, 1)])
def test_invalid_metric_type(tmp_dir, invalid_type):
    dvclive = Live()

    with pytest.raises(
        InvalidDataTypeError,
        match=f"Data 'm' has not supported type {type(invalid_type)}",
    ):
        dvclive.log_metric("m", invalid_type)


def test_get_step_resume(tmp_dir):
    dvclive = Live()

    for metric in [0.9, 0.8]:
        dvclive.log_metric("metric", metric)
        dvclive.next_step()

    assert dvclive.step == 2

    dvclive = Live(resume=True)
    assert dvclive.step == 2

    dvclive = Live(resume=False)
    assert dvclive.step == 0


def test_get_step_custom_steps(tmp_dir):
    dvclive = Live()

    steps = [0, 62, 1000]
    metrics = [0.9, 0.8, 0.7]

    for step, metric in zip(steps, metrics):
        dvclive.step = step
        dvclive.log_metric("x", metric)
        assert dvclive.step == step


def test_get_step_control_flow(tmp_dir):
    dvclive = Live()

    while dvclive.step < 10:
        dvclive.log_metric("i", dvclive.step)
        dvclive.next_step()

    steps, values = read_history(dvclive, "i")
    assert steps == list(range(10))
    assert values == [float(x) for x in range(10)]


def test_logger(tmp_dir, mocker, monkeypatch):
    logger = mocker.patch("dvclive.live.logger")

    live = Live()
    live.log_metric("foo", 0)
    logger.debug.assert_called_with("Logged foo: 0")
    live.next_step()
    logger.debug.assert_called_with("Step: 1")
    live.log_metric("foo", 1)
    live.next_step()

    live = Live(resume=True)
    logger.info.assert_called_with("Resuming from step 1")


def test_make_summary_without_calling_log(tmp_dir):
    dvclive = Live()

    dvclive.summary["foo"] = 1.0
    dvclive.make_summary()

    assert json.loads((tmp_dir / dvclive.metrics_file).read_text()) == {
        # no `step`
        "foo": 1.0
    }
    log_file = tmp_dir / dvclive.plots_dir / Metric.subfolder / "foo.tsv"
    assert not log_file.exists()


@pytest.mark.parametrize("timestamp", [True, False])
def test_log_metric_timestamp(tmp_dir, timestamp):
    live = Live()
    live.log_metric("foo", 1.0, timestamp=timestamp)
    live.next_step()

    history, _ = parse_metrics(live)
    logged = next(iter(history.values()))
    assert ("timestamp" in logged[0]) == timestamp


def test_make_summary_is_called_on_end(tmp_dir):
    live = Live()

    live.summary["foo"] = 1.0
    live.end()

    assert json.loads((tmp_dir / live.metrics_file).read_text()) == {
        # no `step`
        "foo": 1.0
    }
    log_file = tmp_dir / live.plots_dir / Metric.subfolder / "foo.tsv"
    assert not log_file.exists()


def test_make_summary_on_end_dont_increment_step(tmp_dir):
    with Live() as live:
        for i in range(2):
            live.log_metric("foo", i)
            live.next_step()

    assert json.loads((tmp_dir / live.metrics_file).read_text()) == {
        "foo": 1.0,
        "step": 1,
    }


def test_context_manager(tmp_dir):
    with Live() as live:
        live.summary["foo"] = 1.0

    assert json.loads((tmp_dir / live.metrics_file).read_text()) == {
        # no `step`
        "foo": 1.0
    }
    log_file = tmp_dir / live.plots_dir / Metric.subfolder / "foo.tsv"
    assert not log_file.exists()
    report_file = tmp_dir / live.report_file
    assert report_file.exists()


def test_context_manager_skips_end_calls(tmp_dir):
    with Live() as live:
        live.summary["foo"] = 1.0
        live.end()
        assert not (tmp_dir / live.metrics_file).exists()
    assert (tmp_dir / live.metrics_file).exists()


@pytest.mark.parametrize("dvc_root", [True, False])
@pytest.mark.parametrize("set_env", [True, False])
def test_create_checkpoint_file(tmp_dir, monkeypatch, dvc_root, set_env, mocker):
    if dvc_root:
        (tmp_dir / ".dvc" / "tmp").mkdir(parents=True)
    if set_env:
        monkeypatch.setenv(env.DVC_CHECKPOINT, "1")

    dvclive = Live()

    native_exists = os.path.exists

    def checkpoint_never_exists(path):
        # we need that to avoid waiting for dvc process to remove file
        if path.endswith(env.DVC_CHECKPOINT):
            return False

        return native_exists(path)

    with mocker.patch(
        "dvclive.dvc.os.path.exists", side_effect=checkpoint_never_exists
    ):
        dvclive.make_checkpoint()

    if dvc_root and set_env:
        assert native_exists(os.path.join(".dvc", "tmp", env.DVC_CHECKPOINT))
    else:
        assert not native_exists(os.path.join(".dvc", "tmp", env.DVC_CHECKPOINT))


@pytest.mark.vscode()
@pytest.mark.parametrize("dvc_root", [True, False])
def test_vscode_dvclive_only_signal_file(tmp_dir, dvc_root, mocker):
    signal_file = os.path.join(tmp_dir, ".dvc", "tmp", "exps", "run", "DVCLIVE_ONLY")
    test_pid = 12345

    if dvc_root:
        (tmp_dir / ".dvc").mkdir(parents=True)

    assert not os.path.exists(signal_file)

    dvc_repo = mocker.MagicMock()
    dvc_repo.index.stages = []
    dvc_repo.config = {}
    dvc_repo.scm.get_rev.return_value = "current_rev"
    dvc_repo.scm.get_ref.return_value = None
    dvc_repo.scm.no_commits = False
    with mocker.patch("dvclive.live.get_dvc_repo", return_value=dvc_repo), mocker.patch(
        "dvclive.live.os.getpid", return_value=test_pid
    ):
        dvclive = Live(save_dvc_exp=True)

    if dvc_root:
        assert os.path.exists(signal_file)
        with open(signal_file, encoding="utf-8") as f:
            assert f.read() == str(test_pid)

    else:
        assert not os.path.exists(signal_file)

    dvclive.end()

    assert not os.path.exists(signal_file)


@pytest.mark.parametrize(
    "dvcyaml",
    [True, False],
)
def test_make_dvcyaml(tmp_dir, dvcyaml):
    dvclive = Live("logs", dvcyaml=dvcyaml)
    dvclive.log_metric("m1", 1)
    dvclive.make_dvcyaml()

    dvcyaml_path = tmp_dir / dvclive.dir / "dvc.yaml"

    assert dvcyaml_path.is_file()


def test_suppress_dvc_logs(tmp_dir, mocked_dvc_repo):
    Live()
    assert logging.getLogger("dvc").level == 30



import os

import pytest

from dvclive.plots.metric import Metric
from dvclive.serialize import load_yaml
from dvclive.utils import parse_metrics

try:
    import torch
    from pytorch_lightning import LightningModule
    from pytorch_lightning.trainer import Trainer
    from torch import nn
    from torch.nn import functional as F  # noqa: N812
    from torch.optim import SGD, Adam
    from torch.utils.data import DataLoader, Dataset

    from dvclive.lightning import DVCLiveLogger
except ImportError:
    pytest.skip("skipping pytorch_lightning tests", allow_module_level=True)


class XORDataset(Dataset):
    def __init__(self, *args, **kwargs):
        self.ins = [[0, 0], [0, 1], [1, 0], [1, 1]]
        self.outs = [1, 0, 0, 1]

    def __getitem__(self, index):
        return torch.Tensor(self.ins[index]), torch.tensor(
            self.outs[index], dtype=torch.long
        )

    def __len__(self):
        return len(self.ins)


class LitXOR(LightningModule):
    def __init__(
        self, latent_dims=4, optim=SGD, optim_params={"lr": 0.01}  # noqa: B006
    ):
        super().__init__()

        self.save_hyperparameters()

        self.layer_1 = nn.Linear(2, latent_dims)
        self.layer_2 = nn.Linear(latent_dims, 2)

    def forward(self, *args, **kwargs):
        x = args[0]
        batch_size, _ = x.size()
        x = x.view(batch_size, -1)
        x = self.layer_1(x)
        x = F.relu(x)
        x = self.layer_2(x)
        x = F.log_softmax(x, dim=1)
        return x

    def train_loader(self):
        dataset = XORDataset()
        return DataLoader(dataset, batch_size=1)

    def train_dataloader(self):
        return self.train_loader()

    def training_step(self, *args, **kwargs):
        batch = args[0]
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        self.log(
            "train_loss",
            loss,
            prog_bar=True,
            logger=True,
            on_step=True,
            on_epoch=True,
        )
        return loss

    def configure_optimizers(self):
        return self.hparams.optim(self.parameters(), **self.hparams.optim_params)

    def predict_dataloader(self):
        pass

    def test_dataloader(self):
        pass

    def val_dataloader(self):
        pass


def test_lightning_integration(tmp_dir, mocker):
    # init model
    model = LitXOR(latent_dims=8, optim=Adam, optim_params={"lr": 0.02})
    # init logger
    dvclive_logger = DVCLiveLogger("test_run", dir="logs")
    live = dvclive_logger.experiment
    spy = mocker.spy(live, "end")
    trainer = Trainer(
        logger=dvclive_logger,
        max_epochs=2,
        enable_checkpointing=False,
        log_every_n_steps=1,
    )
    trainer.fit(model)
    spy.assert_called_once()

    assert os.path.exists("logs")
    assert not os.path.exists("DvcLiveLogger")

    scalars = os.path.join(dvclive_logger.experiment.plots_dir, Metric.subfolder)
    logs, _ = parse_metrics(dvclive_logger.experiment)

    assert len(logs) == 3
    assert os.path.join(scalars, "train", "epoch", "loss.tsv") in logs
    assert os.path.join(scalars, "train", "step", "loss.tsv") in logs
    assert os.path.join(scalars, "epoch.tsv") in logs

    params_file = dvclive_logger.experiment.params_file
    assert os.path.exists(params_file)
    assert load_yaml(params_file) == {
        "latent_dims": 8,
        "optim": "Adam",
        "optim_params": {"lr": 0.02},
    }


def test_lightning_default_dir(tmp_dir):
    model = LitXOR()
    # If `dir` is not provided handle it properly, use default value
    dvclive_logger = DVCLiveLogger("test_run")
    trainer = Trainer(
        logger=dvclive_logger,
        max_epochs=2,
        enable_checkpointing=False,
        log_every_n_steps=1,
    )
    trainer.fit(model)

    assert os.path.exists("dvclive")


def test_lightning_kwargs(tmp_dir):
    model = LitXOR()
    # Handle kwargs passed to Live.
    dvclive_logger = DVCLiveLogger(dir="dir", report="md", dvcyaml=False)
    trainer = Trainer(
        logger=dvclive_logger,
        max_epochs=2,
        enable_checkpointing=False,
        log_every_n_steps=1,
    )
    trainer.fit(model)

    assert os.path.exists("dir")
    assert os.path.exists("dir/report.md")
    assert not os.path.exists("dir/dvc.yaml")


def test_lightning_steps(tmp_dir, mocker):
    model = LitXOR()
    # Handle kwargs passed to Live.
    dvclive_logger = DVCLiveLogger(dir="logs")
    live = dvclive_logger.experiment
    spy = mocker.spy(live, "next_step")
    trainer = Trainer(
        logger=dvclive_logger,
        max_epochs=2,
        enable_checkpointing=False,
        # Log one time in the middle of the epoch
        log_every_n_steps=3,
    )
    trainer.fit(model)

    history, latest = parse_metrics(dvclive_logger.experiment)
    assert latest["step"] == 7
    assert latest["epoch"] == 1

    scalars = os.path.join(dvclive_logger.experiment.plots_dir, Metric.subfolder)
    epoch_loss = history[os.path.join(scalars, "train", "epoch", "loss.tsv")]
    step_loss = history[os.path.join(scalars, "train", "step", "loss.tsv")]
    assert len(epoch_loss) == 2
    assert len(step_loss) == 2

    # call next_step:
    # - 2x epoch end
    # - 2x log_every_n_steps
    assert spy.call_count == 4


class ValLitXOR(LitXOR):
    def val_loader(self):
        dataset = XORDataset()
        return DataLoader(dataset, batch_size=1)

    def val_dataloader(self):
        return self.val_loader()

    def training_step(self, *args, **kwargs):
        batch = args[0]
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        self.log("train_loss", loss, on_step=True)
        return loss

    def validation_step(self, *args, **kwargs):
        batch = args[0]
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        self.log("val_loss", loss, on_step=False, on_epoch=True)
        return loss


def test_lightning_val_udpates_to_studio(tmp_dir, mocked_dvc_repo, mocked_studio_post):
    """Test the `self.experiment._latest_studio_step -= 1` logic."""
    mocked_post, _ = mocked_studio_post

    model = ValLitXOR()
    dvclive_logger = DVCLiveLogger(save_dvc_exp=True)
    trainer = Trainer(
        logger=dvclive_logger,
        max_steps=4,
        val_check_interval=2,
        log_every_n_steps=1,
        enable_checkpointing=False,
    )
    trainer.fit(model)

    calls = mocked_post.call_args_list
    # 0: start
    # 1: update_train_step_metrics
    # 2: update_train_step_metrics
    # 3: log_eval_end_metrics
    plots = calls[3][1]["json"]["plots"]
    val_loss = plots["dvclive/dvc.yaml::dvclive/plots/metrics/val/loss.tsv"]
    # Without `self.experiment._latest_studio_step -= 1`
    # This would be empty
    assert len(val_loss["data"]) == 1



import json

import numpy as np
import pytest

from dvclive import Live
from dvclive.plots.metric import Metric
from dvclive.plots.utils import NUMPY_INTS, NUMPY_SCALARS
from dvclive.utils import parse_tsv


@pytest.mark.parametrize("dtype", NUMPY_SCALARS)
def test_numpy(tmp_dir, dtype):
    scalar = np.random.rand(1).astype(dtype)[0]
    live = Live()

    live.log_metric("scalar", scalar)
    live.next_step()

    parsed = json.loads((tmp_dir / live.metrics_file).read_text())
    assert isinstance(parsed["scalar"], int if dtype in NUMPY_INTS else float)
    tsv_file = tmp_dir / live.plots_dir / Metric.subfolder / "scalar.tsv"
    tsv_val = parse_tsv(tsv_file)[0]["scalar"]
    assert tsv_val == str(scalar)


def test_name_with_dot(tmp_dir):
    """Regression test for #284"""
    live = Live()

    live.log_metric("scalar.foo.bar", 1.0)
    live.next_step()

    tsv_file = tmp_dir / live.plots_dir / Metric.subfolder / "scalar.foo.bar.tsv"
    assert tsv_file.exists()
    tsv_val = parse_tsv(tsv_file)[0]["scalar.foo.bar"]
    assert tsv_val == "1.0"



import os

import pytest

from dvclive import Live
from dvclive.plots.metric import Metric
from dvclive.utils import parse_metrics

try:
    import numpy as np
    import torch
    from torch import nn
    from transformers import (
        PretrainedConfig,
        PreTrainedModel,
        Trainer,
        TrainingArguments,
    )

    from dvclive.huggingface import DVCLiveCallback
except ImportError:
    pytest.skip("skipping huggingface tests", allow_module_level=True)


def compute_metrics(eval_preds):
    """https://github.com/iterative/dvclive/pull/321#issuecomment-1266916039"""
    import time

    time.sleep(time.get_clock_info("time").resolution)
    return {"foo": 1}


# From transformers/tests/trainer


class RegressionDataset:
    def __init__(self, a=2, b=3, length=64, seed=42, label_names=None):
        np.random.seed(seed)
        self.label_names = ["labels"] if label_names is None else label_names
        self.length = length
        self.x = np.random.normal(size=(length,)).astype(np.float32)
        self.ys = [
            a * self.x + b + np.random.normal(scale=0.1, size=(length,))
            for _ in self.label_names
        ]
        self.ys = [y.astype(np.float32) for y in self.ys]

    def __len__(self):
        return self.length

    def __getitem__(self, i):
        result = {name: y[i] for name, y in zip(self.label_names, self.ys)}
        result["input_x"] = self.x[i]
        return result


class RegressionModelConfig(PretrainedConfig):
    def __init__(self, a=0, b=0, double_output=False, random_torch=True, **kwargs):
        super().__init__(**kwargs)
        self.a = a
        self.b = b
        self.double_output = double_output
        self.random_torch = random_torch
        self.hidden_size = 1


class RegressionPreTrainedModel(PreTrainedModel):
    config_class = RegressionModelConfig
    base_model_prefix = "regression"

    def __init__(self, config):
        super().__init__(config)
        self.a = nn.Parameter(torch.tensor(config.a).float())
        self.b = nn.Parameter(torch.tensor(config.b).float())
        self.double_output = config.double_output

    def forward(self, input_x, labels=None, **kwargs):
        y = input_x * self.a + self.b
        if labels is None:
            return (y, y) if self.double_output else (y,)
        loss = nn.functional.mse_loss(y, labels)
        return (loss, y, y) if self.double_output else (loss, y)


@pytest.fixture()
def data():
    return RegressionDataset(), RegressionDataset()


@pytest.fixture()
def model():
    config = RegressionModelConfig()
    return RegressionPreTrainedModel(config)


@pytest.fixture()
def args():
    return TrainingArguments(
        "foo",
        evaluation_strategy="epoch",
        num_train_epochs=2,
    )


def test_huggingface_integration(tmp_dir, model, args, data, mocker):
    trainer = Trainer(
        model,
        args,
        train_dataset=data[0],
        eval_dataset=data[1],
        compute_metrics=compute_metrics,
    )
    callback = DVCLiveCallback()
    live = callback.live
    spy = mocker.spy(live, "end")
    trainer.add_callback(callback)
    trainer.train()
    spy.assert_called_once()

    live = callback.live
    assert os.path.exists(live.dir)

    logs, _ = parse_metrics(live)

    assert len(logs) == 10

    scalars = os.path.join(live.plots_dir, Metric.subfolder)
    assert os.path.join(scalars, "eval", "foo.tsv") in logs
    assert os.path.join(scalars, "eval", "loss.tsv") in logs
    assert os.path.join(scalars, "train", "loss.tsv") in logs
    assert len(logs[os.path.join(scalars, "epoch.tsv")]) == 3
    assert len(logs[os.path.join(scalars, "eval", "loss.tsv")]) == 2


def test_huggingface_model_file(tmp_dir, model, args, data, mocker):
    model_path = tmp_dir / "model_hf"
    model_save = mocker.spy(model, "save_pretrained")

    live_callback = DVCLiveCallback(model_file=model_path)
    log_artifact = mocker.patch.object(live_callback.live, "log_artifact")

    trainer = Trainer(
        model,
        args,
        train_dataset=data[0],
        eval_dataset=data[1],
        compute_metrics=compute_metrics,
    )
    trainer.add_callback(live_callback)
    trainer.train()

    assert model_path.is_dir()

    assert (model_path / "pytorch_model.bin").exists()
    assert (model_path / "config.json").exists()
    assert model_save.call_count == 2
    log_artifact.assert_called_with(model_path)


def test_huggingface_pass_logger():
    logger = Live("train_logs")

    assert DVCLiveCallback().live is not logger
    assert DVCLiveCallback(live=logger).live is logger


def test_huggingface_log_artifact(tmp_dir, model, args, data, mocker):
    live_callback = DVCLiveCallback()
    log_artifact = mocker.patch.object(live_callback.live, "log_artifact")

    args.load_best_model_at_end = True
    trainer = Trainer(
        model,
        args,
        train_dataset=data[0],
        eval_dataset=data[1],
        compute_metrics=compute_metrics,
    )
    trainer.add_callback(live_callback)
    trainer.train()

    log_artifact.assert_called_with(trainer.args.output_dir)



import inspect
from typing import Optional

from fastai.callback.core import Callback

from dvclive import Live
from dvclive.utils import standardize_metric_name


def _inside_fine_tune():
    """
    Hack to find out if fastai is calling `after_fit` at the end of the
    "freeze" stage part of `learn.fine_tune` .
    """
    fine_tune = False
    fit_one_cycle = False
    for frame in inspect.stack():
        if frame.function == "fine_tune":
            fine_tune = True
        if frame.function == "fit_one_cycle":
            fit_one_cycle = True
        if fine_tune and fit_one_cycle:
            return True
    return False


class DVCLiveCallback(Callback):
    def __init__(
        self,
        model_file: Optional[str] = None,
        with_opt: bool = False,
        live: Optional[Live] = None,
        **kwargs,
    ):
        super().__init__()
        self.model_file = model_file
        self.with_opt = with_opt
        self.live = live if live is not None else Live(**kwargs)
        self.freeze_stage_ended = False

    def before_fit(self):
        if hasattr(self, "lr_finder") or hasattr(self, "gather_preds"):
            return
        params = {
            "model": type(self.learn.model).__qualname__,
            "batch_size": getattr(self.dls, "bs", None),
            "batch_per_epoch": len(getattr(self.dls, "train", [])),
            "frozen": bool(getattr(self.opt, "frozen_idx", -1)),
            "frozen_idx": getattr(self.opt, "frozen_idx", -1),
            "transforms": f"{getattr(self.dls, 'tfms', None)}",
        }
        self.live.log_params(params)

    def after_epoch(self):
        if hasattr(self, "lr_finder") or hasattr(self, "gather_preds"):
            return
        logged_metrics = False
        for key, value in zip(
            self.learn.recorder.metric_names, self.learn.recorder.log
        ):
            if key == "epoch":
                continue
            self.live.log_metric(standardize_metric_name(key, __name__), float(value))
            logged_metrics = True

        # When resuming (i.e. passing `start_epoch` to learner)
        # fast.ai calls after_epoch but we don't want to increase the step.
        if logged_metrics:
            if self.model_file:
                file = self.learn.save(self.model_file, with_opt=self.with_opt)
                self.live.log_artifact(str(file))
            self.live.next_step()

    def after_fit(self):
        if hasattr(self, "lr_finder") or hasattr(self, "gather_preds"):
            return
        if _inside_fine_tune() and not self.freeze_stage_ended:
            self.freeze_stage_ended = True
        else:
            if hasattr(self, "save_model") and self.save_model.last_saved_path:
                self.live.log_artifact(str(self.save_model.last_saved_path))
            self.live.end()



import os

import pytest

from dvclive import Live
from dvclive.plots.metric import Metric

try:
    from fastai.callback.tracker import SaveModelCallback
    from fastai.tabular.all import (
        Categorify,
        Normalize,
        ProgressCallback,
        TabularDataLoaders,
        accuracy,
        tabular_learner,
    )

    from dvclive.fastai import DVCLiveCallback
except ImportError:
    pytest.skip("skipping fastai tests", allow_module_level=True)


@pytest.fixture()
def data_loader():
    from pandas import DataFrame

    d = {
        "x1": [1, 1, 0, 0, 1, 1, 0, 0],
        "x2": [1, 0, 1, 0, 1, 0, 1, 0],
        "y": [1, 0, 0, 1, 1, 0, 0, 1],
    }
    df = DataFrame(d)
    return TabularDataLoaders.from_df(
        df,
        valid_idx=[4, 5, 6, 7],
        batch_size=2,
        cont_names=["x1", "x2"],
        procs=[Categorify, Normalize],
        y_names="y",
    )


def test_fastai_callback(tmp_dir, data_loader, mocker):
    learn = tabular_learner(data_loader, metrics=accuracy)
    learn.remove_cb(ProgressCallback)
    callback = DVCLiveCallback()
    live = callback.live

    spy = mocker.spy(live, "end")
    learn.fit_one_cycle(2, cbs=[callback])
    spy.assert_called_once()

    assert (tmp_dir / live.dir).exists()
    assert (tmp_dir / live.params_file).exists()
    assert (tmp_dir / live.params_file).read_text() == (
        "model: TabularModel\nbatch_size: 2\nbatch_per_epoch: 2\nfrozen: false"
        "\nfrozen_idx: 0\ntransforms: None\n"
    )

    metrics_path = tmp_dir / live.plots_dir / Metric.subfolder
    train_path = metrics_path / "train"
    valid_path = metrics_path / "eval"

    assert train_path.is_dir()
    assert valid_path.is_dir()
    assert (metrics_path / "accuracy.tsv").exists()
    assert not (metrics_path / "epoch.tsv").exists()


def test_fastai_model_file(tmp_dir, data_loader, mocker):
    learn = tabular_learner(data_loader, metrics=accuracy)
    learn.remove_cb(ProgressCallback)
    learn.model_dir = os.path.abspath("./")
    save = mocker.spy(learn, "save")
    live_callback = DVCLiveCallback("model", with_opt=True)
    log_artifact = mocker.patch.object(live_callback.live, "log_artifact")
    learn.fit_one_cycle(2, cbs=[live_callback])
    assert (tmp_dir / "model.pth").is_file()
    save.assert_called_with("model", with_opt=True)
    log_artifact.assert_called_with(str(tmp_dir / "model.pth"))


def test_fastai_pass_logger():
    logger = Live("train_logs")

    assert DVCLiveCallback().live is not logger
    assert DVCLiveCallback(live=logger).live is logger


def test_fast_ai_resume(tmp_dir, data_loader, mocker):
    learn = tabular_learner(data_loader, metrics=accuracy)
    learn.remove_cb(ProgressCallback)
    callback = DVCLiveCallback()
    live = callback.live

    spy = mocker.spy(live, "next_step")
    end = mocker.spy(live, "end")
    learn.fit_one_cycle(2, cbs=[callback])
    assert spy.call_count == 2
    assert end.call_count == 1

    callback = DVCLiveCallback(resume=True)
    live = callback.live
    spy = mocker.spy(live, "next_step")
    learn.fit_one_cycle(3, cbs=[callback], start_epoch=live.step)
    assert spy.call_count == 1


def test_fast_ai_avoid_unnecessary_end_calls(tmp_dir, data_loader, mocker):
    """
    `after_fit` might be called from different points and not all mean that the
    training has ended.
    """
    learn = tabular_learner(data_loader, metrics=accuracy)
    learn.remove_cb(ProgressCallback)
    callback = DVCLiveCallback()
    live = callback.live

    end = mocker.spy(live, "end")
    after_fit = mocker.spy(callback, "after_fit")
    learn.fine_tune(2, cbs=[callback])
    assert end.call_count == 1
    assert after_fit.call_count == 2


def test_fastai_save_model_callback(tmp_dir, data_loader, mocker):
    learn = tabular_learner(data_loader, metrics=accuracy)
    learn.remove_cb(ProgressCallback)
    learn.model_dir = os.path.abspath("./")

    save_callback = SaveModelCallback()
    live_callback = DVCLiveCallback()
    log_artifact = mocker.patch.object(live_callback.live, "log_artifact")
    learn.fit_one_cycle(2, cbs=[save_callback, live_callback])
    assert (tmp_dir / "model.pth").is_file()
    log_artifact.assert_called_with(str(save_callback.last_saved_path))



import json
import logging
import os
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from dvc_studio_client.post_live_metrics import post_live_metrics
from funcy import set_in
from pathspec import PathSpec
from ruamel.yaml.representer import RepresenterError

from . import env
from .dvc import (
    get_dvc_repo,
    get_random_exp_name,
    make_checkpoint,
    make_dvcyaml,
    mark_dvclive_only_ended,
    mark_dvclive_only_started,
)
from .error import (
    InvalidDataTypeError,
    InvalidParameterTypeError,
    InvalidPlotTypeError,
    InvalidReportModeError,
)
from .plots import PLOT_TYPES, SKLEARN_PLOTS, CustomPlot, Image, Metric, NumpyEncoder
from .report import BLANK_NOTEBOOK_REPORT, make_report
from .serialize import dump_json, dump_yaml, load_yaml
from .studio import get_dvc_studio_config, get_studio_updates
from .utils import (
    StrPath,
    clean_and_copy_into,
    env2bool,
    inside_notebook,
    matplotlib_installed,
    open_file_in_browser,
)

logger = logging.getLogger("dvclive")
handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(os.getenv(env.DVCLIVE_LOGLEVEL, "INFO").upper())

ParamLike = Union[int, float, str, bool, List["ParamLike"], Dict[str, "ParamLike"]]


class Live:
    def __init__(
        self,
        dir: str = "dvclive",  # noqa: A002
        resume: bool = False,
        report: Optional[str] = "auto",
        save_dvc_exp: bool = False,
        dvcyaml: bool = True,
        exp_message: Optional[str] = None,
    ):
        self.summary: Dict[str, Any] = {}

        self._dir: str = dir
        self._resume: bool = resume or env2bool(env.DVCLIVE_RESUME)
        self._save_dvc_exp: bool = save_dvc_exp
        self._step: Optional[int] = None
        self._metrics: Dict[str, Any] = {}
        self._images: Dict[str, Any] = {}
        self._params: Dict[str, Any] = {}
        self._plots: Dict[str, Any] = {}
        self._artifacts: Dict[str, Dict] = {}
        self._inside_with = False
        self._dvcyaml = dvcyaml

        os.makedirs(self.dir, exist_ok=True)

        self._report_mode: Optional[str] = report
        self._report_notebook = None
        self._init_report()

        if self._resume:
            self._init_resume()
        else:
            self._init_cleanup()

        self._baseline_rev: Optional[str] = None
        self._exp_name: Optional[str] = None
        self._exp_message: Optional[str] = exp_message
        self._experiment_rev: Optional[str] = None
        self._inside_dvc_exp: bool = False
        self._dvc_repo = None
        self._include_untracked: List[str] = []
        self._init_dvc()

        self._latest_studio_step = self.step if resume else -1
        self._studio_events_to_skip: Set[str] = set()
        self._dvc_studio_config: Dict[str, Any] = {}
        self._init_studio()

    def _init_resume(self):
        self._read_params()
        self._step = self.read_step()
        if self._step != 0:
            logger.info(f"Resuming from step {self._step}")
            self._step += 1
        logger.debug(f"{self._step=}")

    def _init_cleanup(self):
        for plot_type in PLOT_TYPES:
            shutil.rmtree(
                Path(self.plots_dir) / plot_type.subfolder, ignore_errors=True
            )

        for f in (
            self.metrics_file,
            self.report_file,
            self.params_file,
        ):
            if f and os.path.exists(f):
                os.remove(f)

    def _init_dvc(self):
        from dvc.scm import NoSCM

        if os.getenv(env.DVC_EXP_BASELINE_REV, None):
            # `dvc exp` execution
            self._baseline_rev = os.getenv(env.DVC_EXP_BASELINE_REV, "")
            self._exp_name = os.getenv(env.DVC_EXP_NAME, "")
            self._inside_dvc_exp = True
            if self._save_dvc_exp:
                logger.warning(
                    "Ignoring `_save_dvc_exp` because `dvc exp run` is running"
                )
                self._save_dvc_exp = False

        self._dvc_repo = get_dvc_repo()

        dvc_logger = logging.getLogger("dvc")
        dvc_logger.setLevel(os.getenv(env.DVCLIVE_LOGLEVEL, "WARNING").upper())

        if (self._dvc_repo is None) or isinstance(self._dvc_repo.scm, NoSCM):
            if self._save_dvc_exp:
                logger.warning(
                    "Can't save experiment without a Git Repo."
                    "\nCreate a Git repo (`git init`) and commit (`git commit`)."
                )
                self._save_dvc_exp = False
            return
        if self._dvc_repo.scm.no_commits:
            if self._save_dvc_exp:
                logger.warning(
                    "Can't save experiment to an empty Git Repo."
                    "\nAdd a commit (`git commit`) to save experiments."
                )
                self._save_dvc_exp = False
            return

        if self._inside_dvc_exp:
            return

        self._baseline_rev = self._dvc_repo.scm.get_rev()
        if self._save_dvc_exp:
            self._exp_name = get_random_exp_name(self._dvc_repo.scm, self._baseline_rev)
            mark_dvclive_only_started()
            self._include_untracked.append(self.dir)

    def _init_studio(self):
        self._dvc_studio_config = get_dvc_studio_config(self)
        if not self._dvc_studio_config:
            logger.debug("Skipping `studio` report.")
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("data")
            self._studio_events_to_skip.add("done")
        elif self._inside_dvc_exp:
            logger.debug("Skipping `studio` report `start` and `done` events.")
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("done")
        elif self._dvc_repo is None:
            logger.warning(
                "Can't connect to Studio without a DVC Repo."
                "\nYou can create a DVC Repo by calling `dvc init`."
            )
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("data")
            self._studio_events_to_skip.add("done")
        elif not self._save_dvc_exp:
            logger.warning(
                "Can't connect to Studio without creating a DVC experiment."
                "\nIf you have a DVC Pipeline, run it with `dvc exp run`."
                "\nIf you are using DVCLive alone, use `save_dvc_exp=True`."
            )
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("data")
            self._studio_events_to_skip.add("done")
        else:
            response = post_live_metrics(
                "start",
                self._baseline_rev,
                self._exp_name,
                "dvclive",
                dvc_studio_config=self._dvc_studio_config,
                message=self._exp_message,
            )
            if not response:
                logger.debug(
                    "`studio` report `start` event failed. "
                    "`studio` report cancelled."
                )
                self._studio_events_to_skip.add("start")
                self._studio_events_to_skip.add("data")
                self._studio_events_to_skip.add("done")

    def _init_report(self):
        if self._report_mode == "auto":
            if env2bool("CI") and matplotlib_installed():
                self._report_mode = "md"
            else:
                self._report_mode = "html"
        elif self._report_mode == "notebook":
            if inside_notebook():
                from IPython.display import Markdown, display

                self._report_mode = "notebook"
                self._report_notebook = display(
                    Markdown(BLANK_NOTEBOOK_REPORT), display_id=True
                )
            else:
                self._report_mode = "html"
        elif self._report_mode not in {None, "html", "notebook", "md"}:
            raise InvalidReportModeError(self._report_mode)
        logger.debug(f"{self._report_mode=}")

    @property
    def dir(self) -> str:  # noqa: A003
        return self._dir

    @property
    def params_file(self) -> str:
        return os.path.join(self.dir, "params.yaml")

    @property
    def metrics_file(self) -> str:
        return os.path.join(self.dir, "metrics.json")

    @property
    def dvc_file(self) -> str:
        return os.path.join(self.dir, "dvc.yaml")

    @property
    def plots_dir(self) -> str:
        return os.path.join(self.dir, "plots")

    @property
    def artifacts_dir(self) -> str:
        return os.path.join(self.dir, "artifacts")

    @property
    def report_file(self) -> Optional[str]:
        if self._report_mode in ("html", "md"):
            suffix = self._report_mode
            return os.path.join(self.dir, f"report.{suffix}")
        return None

    @property
    def step(self) -> int:
        return self._step or 0

    @step.setter
    def step(self, value: int) -> None:
        self._step = value
        logger.debug(f"Step: {self.step}")

    def next_step(self):
        if self._step is None:
            self._step = 0

        self.make_summary()

        if self._dvcyaml:
            self.make_dvcyaml()

        self.make_report()
        self.make_checkpoint()
        self.step += 1

    def log_metric(
        self,
        name: str,
        val: Union[int, float],
        timestamp: bool = False,
        plot: bool = True,
    ):
        if not Metric.could_log(val):
            raise InvalidDataTypeError(name, type(val))

        if name in self._metrics:
            metric = self._metrics[name]
        else:
            metric = Metric(name, self.plots_dir)
            self._metrics[name] = metric

        metric.step = self.step
        if plot:
            metric.dump(val, timestamp=timestamp)

        self.summary = set_in(self.summary, metric.summary_keys, val)
        logger.debug(f"Logged {name}: {val}")

    def log_image(self, name: str, val):
        if not Image.could_log(val):
            raise InvalidDataTypeError(name, type(val))

        if isinstance(val, (str, Path)):
            from PIL import Image as ImagePIL

            val = ImagePIL.open(val)

        if name in self._images:
            image = self._images[name]
        else:
            image = Image(name, self.plots_dir)
            self._images[name] = image

        image.step = self.step
        image.dump(val)
        logger.debug(f"Logged {name}: {val}")

    def log_plot(
        self,
        name: str,
        datapoints: List[Dict],
        x: str,
        y: str,
        template: Optional[str] = None,
        title: Optional[str] = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
    ):
        if not CustomPlot.could_log(datapoints):
            raise InvalidDataTypeError(name, type(datapoints))

        if name in self._plots:
            plot = self._plots[name]
        else:
            plot = CustomPlot(
                name,
                self.plots_dir,
                x=x,
                y=y,
                template=template,
                title=title,
                x_label=x_label,
                y_label=y_label,
            )
            self._plots[name] = plot

        plot.step = self.step
        plot.dump(datapoints)
        logger.debug(f"Logged {name}")

    def log_sklearn_plot(self, kind, labels, predictions, name=None, **kwargs):
        val = (labels, predictions)

        plot_config = {
            k: v
            for k, v in kwargs.items()
            if k in ("title", "x_label", "y_label", "normalized")
        }
        name = name or kind
        if name in self._plots:
            plot = self._plots[name]
        elif kind in SKLEARN_PLOTS and SKLEARN_PLOTS[kind].could_log(val):
            plot = SKLEARN_PLOTS[kind](name, self.plots_dir, **plot_config)
            self._plots[plot.name] = plot
        else:
            raise InvalidPlotTypeError(name)

        sklearn_kwargs = {
            k: v for k, v in kwargs.items() if k not in plot_config or k != "normalized"
        }
        plot.step = self.step
        plot.dump(val, **sklearn_kwargs)
        logger.debug(f"Logged {name}")

    def _read_params(self):
        if os.path.isfile(self.params_file):
            params = load_yaml(self.params_file)
            self._params.update(params)

    def _dump_params(self):
        try:
            dump_yaml(self._params, self.params_file)
        except RepresenterError as exc:
            raise InvalidParameterTypeError(exc.args) from exc

    def log_params(self, params: Dict[str, ParamLike]):
        """Saves the given set of parameters (dict) to yaml"""
        self._params.update(params)
        self._dump_params()
        logger.debug(f"Logged {params} parameters to {self.params_file}")

    def log_param(self, name: str, val: ParamLike):
        """Saves the given parameter value to yaml"""
        self.log_params({name: val})

    def log_artifact(
        self,
        path: StrPath,
        type: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        desc: Optional[str] = None,  # noqa: ARG002
        labels: Optional[List[str]] = None,  # noqa: ARG002
        meta: Optional[Dict[str, Any]] = None,  # noqa: ARG002
        copy: bool = False,
    ):
        """Tracks a local file or directory with DVC"""
        if not isinstance(path, (str, Path)):
            raise InvalidDataTypeError(path, type(path))

        if self._dvc_repo is not None:
            from dvc.repo.artifacts import name_is_compatible

            if copy:
                path = clean_and_copy_into(path, self.artifacts_dir)

            try:
                stage = self._dvc_repo.add(path)
            except Exception as e:  # noqa: BLE001
                logger.warning(f"Failed to dvc add {path}: {e}")
                return

            name = name or Path(path).stem
            if name_is_compatible(name):
                self._artifacts[name] = {
                    k: v
                    for k, v in locals().items()
                    if k in ("path", "type", "desc", "labels", "meta") and v is not None
                }
            else:
                logger.warning(
                    "Can't use '%s' as artifact name (ID)."
                    " It will not be included in the `artifacts` section.",
                    name,
                )

            dvc_file = stage[0].addressing

            if self._save_dvc_exp:
                self._include_untracked.append(dvc_file)
                self._include_untracked.append(
                    str(Path(dvc_file).parent / ".gitignore")
                )

    def make_summary(self, update_step: bool = True):
        if self._step is not None and update_step:
            self.summary["step"] = self.step
        dump_json(self.summary, self.metrics_file, cls=NumpyEncoder)

    def make_report(self):
        if "data" not in self._studio_events_to_skip:
            response = False
            if post_live_metrics is not None:
                metrics, params, plots = get_studio_updates(self)
                response = post_live_metrics(
                    "data",
                    self._baseline_rev,
                    self._exp_name,
                    "dvclive",
                    step=self.step,
                    metrics=metrics,
                    params=params,
                    plots=plots,
                    dvc_studio_config=self._dvc_studio_config,
                )
            if not response:
                logger.warning(
                    "`post_to_studio` `data` event failed."
                    " Data will be resent on next call."
                )
            else:
                self._latest_studio_step = self.step

        if self._report_mode is not None:
            make_report(self)
            if self._report_mode == "html" and env2bool(env.DVCLIVE_OPEN):
                open_file_in_browser(self.report_file)

    def make_dvcyaml(self):
        make_dvcyaml(self)

    def end(self):
        if self._inside_with:
            # Prevent `live.end` calls inside context manager
            return
        self.make_summary(update_step=False)
        if self._dvcyaml:
            self.make_dvcyaml()

        self._ensure_paths_are_tracked_in_dvc_exp()

        self.save_dvc_exp()

        if "done" not in self._studio_events_to_skip:
            response = False
            if post_live_metrics is not None:
                kwargs = {}
                if self._experiment_rev:
                    kwargs["experiment_rev"] = self._experiment_rev
                response = post_live_metrics(
                    "done",
                    self._baseline_rev,
                    self._exp_name,
                    "dvclive",
                    dvc_studio_config=self._dvc_studio_config,
                    **kwargs,
                )
            if not response:
                logger.warning("`post_to_studio` `done` event failed.")
            self._studio_events_to_skip.add("done")
            self._studio_events_to_skip.add("data")
        else:
            self.make_report()

    def make_checkpoint(self):
        if env2bool(env.DVC_CHECKPOINT):
            make_checkpoint()

    def read_step(self):
        if Path(self.metrics_file).exists():
            latest = self.read_latest()
            return latest.get("step", 0)
        return 0

    def read_latest(self):
        with open(self.metrics_file, encoding="utf-8") as fobj:
            return json.load(fobj)

    def __enter__(self):
        self._inside_with = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._inside_with = False
        self.end()

    def _ensure_paths_are_tracked_in_dvc_exp(self):
        if self._inside_dvc_exp and self._dvc_repo:
            dir_spec = PathSpec.from_lines("gitwildmatch", [self.dir])
            outs_spec = PathSpec.from_lines(
                "gitwildmatch", [str(o) for o in self._dvc_repo.index.outs]
            )
            try:
                paths_to_track = [
                    f
                    for f in self._dvc_repo.scm.untracked_files()
                    if (dir_spec.match_file(f) and not outs_spec.match_file(f))
                ]
                if paths_to_track:
                    self._dvc_repo.scm.add(paths_to_track)
            except Exception as e:  # noqa: BLE001
                logger.warning(f"Failed to git add paths:\n{e}")

    def save_dvc_exp(self):
        if self._save_dvc_exp:
            from dvc.exceptions import DvcException

            try:
                self._experiment_rev = self._dvc_repo.experiments.save(
                    name=self._exp_name,
                    include_untracked=self._include_untracked,
                    force=True,
                    message=self._exp_message,
                )
            except DvcException as e:
                logger.warning(f"Failed to save experiment:\n{e}")
            finally:
                mark_dvclive_only_ended()



# ruff: noqa: ARG002
import inspect
from typing import Any, Dict, Optional

from lightning_fabric.utilities.logger import (
    _convert_params,
    _sanitize_callable_params,
    _sanitize_params,
)
from pytorch_lightning.loggers.logger import Logger, rank_zero_experiment
from pytorch_lightning.utilities import rank_zero_only
from torch import is_tensor

from dvclive import Live
from dvclive.utils import standardize_metric_name


def _should_call_next_step():
    """
    Find out if pytorch_lightning is calling `log_metrics` from the functions
    where we actually want to call `next_step`.
    For example, prevents calling next_step when external callbacks call
    `log_metrics` or during the multiple `update_eval_step_metrics`.
    """
    return any(
        frame.function
        in (
            "update_train_step_metrics",
            "update_train_epoch_metrics",
            "log_eval_end_metrics",
        )
        for frame in inspect.stack()
    )


class DVCLiveLogger(Logger):
    def __init__(
        self,
        run_name: Optional[str] = "dvclive_run",
        prefix="",
        experiment=None,
        dir: Optional[str] = None,  # noqa: A002
        resume: bool = False,
        report: Optional[str] = "auto",
        save_dvc_exp: bool = False,
        dvcyaml: bool = True,
    ):
        super().__init__()
        self._prefix = prefix
        self._live_init: Dict[str, Any] = {
            "resume": resume,
            "report": report,
            "save_dvc_exp": save_dvc_exp,
            "dvcyaml": dvcyaml,
        }
        if dir is not None:
            self._live_init["dir"] = dir
        self._experiment = experiment
        self._version = run_name
        # Force Live instantiation
        self.experiment  # noqa: B018

    @property
    def name(self):
        return "DvcLiveLogger"

    @rank_zero_only
    def log_hyperparams(self, params, *args, **kwargs):
        def sanitize_dict(params):
            dict_values = {}
            non_dict_values = {}
            for k, v in params.items():
                if isinstance(v, dict):
                    dict_values[k] = sanitize_dict(v)
                else:
                    non_dict_values[k] = v
            non_dict_values = _sanitize_params(non_dict_values)
            return {**dict_values, **non_dict_values}

        params = _convert_params(params)
        params = _sanitize_callable_params(params)
        params = sanitize_dict(params)
        self.experiment.log_params(params)

    @property  # type: ignore
    @rank_zero_experiment
    def experiment(self):
        r"""
        Actual DVCLive object. To use DVCLive features in your
        :class:`~LightningModule` do the following.
        Example::
            self.logger.experiment.some_dvclive_function()
        """
        if self._experiment is not None:
            return self._experiment
        self._experiment = Live(**self._live_init)

        return self._experiment

    @property
    def version(self):
        return self._version

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, Any], step: Optional[int] = None):
        self.experiment.step = step
        for metric_name, metric_val in metrics.items():
            val = metric_val
            if is_tensor(val):
                val = val.cpu().detach().item()
            name = standardize_metric_name(metric_name, __name__)
            self.experiment.log_metric(name=name, val=val)
        if _should_call_next_step():
            if step == self.experiment._latest_studio_step:  # noqa: SLF001
                # We are in log_eval_end_metrics but there has been already
                # a studio request sent with `step`.
                # We decrease the number to bypass `live.studio._get_unsent_datapoints`
                self.experiment._latest_studio_step -= 1  # noqa: SLF001
            self.experiment.next_step()

    @rank_zero_only
    def finalize(self, status: str) -> None:
        self.experiment.end()



# ruff: noqa: ARG002
from typing import Optional

from catalyst import utils
from catalyst.core.callback import Callback, CallbackOrder

from dvclive import Live


class DVCLiveCallback(Callback):
    def __init__(self, model_file=None, live: Optional[Live] = None, **kwargs):
        super().__init__(order=CallbackOrder.external)
        self.model_file = model_file
        self.live = live if live is not None else Live(**kwargs)

    def on_epoch_end(self, runner) -> None:
        for loader_key, per_loader_metrics in runner.epoch_metrics.items():
            for key, value in per_loader_metrics.items():
                self.live.log_metric(
                    f"{loader_key}/{key.replace('/', '_')}", float(value)
                )

        if self.model_file:
            checkpoint = utils.pack_checkpoint(
                model=runner.model,
                criterion=runner.criterion,
                optimizer=runner.optimizer,
                scheduler=runner.scheduler,
            )
            utils.save_checkpoint(checkpoint, self.model_file)
        self.live.next_step()

    def on_experiment_end(self, runner):
        self.live.end()
